import os
import re
import sys
import numpy as np

from os.path import join as opj

from ...core.basejob import SingleJob
from ...core.errors import PlamsError, ResultsError, FileError, JobError
from ...core.functions import log, parse_heredoc
from ...core.private import sha256, UpdateSysPath
from ...core.results import Results
from ...core.settings import Settings
from ...mol.molecule import Molecule
from ...mol.atom import Atom
from ...tools.kftools import KFFile
from ...tools.units import Units
from .ams import AMSJob



class SCMResults(Results):
    """Abstract class gathering common mechanisms for results of ADF Suite programs."""
    _kfext = ''


    def collect(self):
        """Collect files present in the job folder. Use parent method from |Results|, then create an instance of |KFFile| for the main KF file and store it as ``_kf`` attribute.
        """
        Results.collect(self)
        kfname = self.job.name + self.__class__._kfext
        if kfname in self.files:
            self._kf = KFFile(opj(self.job.path, kfname))
        else:
            log('WARNING: Main KF file {} not present in {}'.format(kfname, self.job.path), 1)


    def refresh(self):
        """Refresh the contents of ``files`` list. Use parent method from |Results|, then look at all attributes that are instances of |KFFile| and check if they point to existing files. If not, try to reinstantiate them with current job path (that can happen while loading a pickled job after the entire job folder was moved).
        """
        Results.refresh(self)
        to_remove = []
        for attr,val in self.__dict__.items():
            if isinstance(val, KFFile) and os.path.dirname(val.path) != self.job.path:
                guessnewpath = opj(self.job.path, os.path.basename(val.path))
                if os.path.isfile(guessnewpath):
                    self.__dict__[attr] = KFFile(guessnewpath)
                else:
                    to_remove.append(attr)
        for i in to_remove:
            del self.__dict__[i]


    def readkf(self, section, variable):
        """readkf(section, variable)
        Read data from *section*/*variable* of the main KF file.

        The type of the returned value depends on the type of *variable* defined inside KF file. It can be: single int, list of ints, single float, list of floats, single boolean, list of booleans or string. """
        if self._kfpresent():
            return self._kf.read(section, variable)
        raise FileError('File {} not present in {}'.format(self.job.name+self.__class__._kfext, self.job.path))


    def newkf(self, filename):
        """newkf(filename)
        Create new |KFFile| instance using file *filename* in the job folder.

        Example usage::

            >>> res = someadfjob.run()
            >>> tape13 = res.newkf('$JN.t13')
            >>> print(tape13.read('Geometry', 'xyz'))

        """
        self.refresh()
        filename = filename.replace('$JN', self.job.name)
        if filename in self.files:
            return KFFile(opj(self.job.path, filename))
        else:
            raise FileError('File {} not present in {}'.format(filename, self.job.path))

    def get_properties(self):
        """get_properties()
        Return a dictionary with all the entries from ``Properties`` section in the main KF file.
        """
        n = self.readkf('Properties', 'nEntries')
        ret = {}
        for i in range(1, n+1):
            tp = self.readkf('Properties', 'Type({})'.format(i)).strip()
            stp = self.readkf('Properties', 'Subtype({})'.format(i)).strip()
            val = self.readkf('Properties', 'Value({})'.format(i))
            key = stp if stp.endswith(tp) else ('{} {}'.format(stp, tp) if stp else tp)
            ret[key] = val
        return ret

    def get_molecule(self, section, variable, unit='bohr', internal=False, n=1):
        """get_molecule(section, variable, unit='bohr', internal=False, n=1)
        Read molecule coordinates from *section*/*variable* of the main KF file.

        Returned |Molecule| instance is created by copying a molecule from associated |SCMJob| instance and updating atomic coordinates with values read from *section*/*variable*. The format in which coordinates are stored is not consistent for all programs or even for different sections of the same KF file. Sometimes coordinates are stored in bohr, sometimes in angstrom. The order of atoms can be either input order or internal order. These settings can be adjusted with *unit* and *internal* parameters. Some variables store more than one geometry, in those cases *n* can be used to choose the preferred one.
        """
        atnums = self._atomic_numbers_input_order()
        natoms = len(atnums)
        coords = self.readkf(section, variable)
        coords = [coords[i:i+3] for i in range(0,len(coords),3)]
        if len(coords) > natoms:
            if len(coords) < n*natoms:
                raise ResultsError('get_molecule() failed. Not enough data in {}%{} to extract geometry no {}'.format(section, variable, n))
            coords = coords[(n-1)*natoms : n*natoms]
        if internal:
            mapping = self._int2inp()
            coords = [coords[mapping[i]-1] for i in range(len(coords))]
        ret = Molecule()
        for z,crd in zip(atnums,coords):
            ret.add_atom(Atom(atnum=z, coords=crd, unit=unit))
        return ret


    def _get_single_value(self, section, variable, output_unit, native_unit='au'):
        """_get_single_value(section, variable, output_unit, native_unit='au')

        A small method template for all the single number "get_something()" methods extracting data from main KF file. Returned value is converted from *native_unit* to *output_unit*.
        """
        if self._kfpresent() and (section, variable) in self._kf:
            return Units.convert(self.readkf(section, variable), native_unit, output_unit)
        raise ResultsError("'{}%{}' not present in {}".format(section, variable, self._kfpath()))


    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order. Abstract method.
        """
        raise PlamsError('Trying to run an abstract method SCMResults._atomic_numbers_input_order()')


    def _kfpath(self):
        """_kfpath()
        Return the absolute path to the main KF file.
        """
        return opj(self.job.path, self.job.name + self.__class__._kfext)


    def _kfpresent(self):
        """_kfpresent()
        Check if this instance has a valid ``_kf`` attribute.
        """
        return hasattr(self, '_kf') and isinstance(self._kf, KFFile)


    def _export_attribute(self, attr, other):
        """_export_attribute(attr, other)
        If *attr* is a KF file take care of a proper path. Otherwise use parent method. See :meth:`Results._copy_to<scm.plams.core.results.Results._copy_to>` for details.
        """
        if isinstance(attr, KFFile):
            oldname = os.path.basename(attr.path)
            newname = Results._replace_job_name(oldname, self.job.name, other.job.name)
            newpath = opj(other.job.path, newname)
            return KFFile(newpath) if os.path.isfile(newpath) else None
        else:
            return Results._export_attribute(self, attr, other)


    def _int2inp(self):
        """_int2inp()
        Obtain mapping from internal atom order to the input one. Abstract method.
        """
        raise PlamsError('Trying to run an abstract method SCMResults._int2inp()')


    def to_input_order(self, data):
        """to_input_order(self, data)
        Reorder any iterable *data* from the internal atom order to the input atom order. The length of *data* must be equal to the number of atoms, otherwise an exception is raised. Returned value is a container of the same type as *data*.
        """
        mapping = self._int2inp()
        if len(mapping) != len(data):
            raise PlamsError('to_input_order() got an argument with incorrect length. Length must be equal to the number of atoms')
        t = np.array if type(data) == np.ndarray else type(data)
        return t([data[mapping[i]-1] for i in range(len(mapping))])


    def readarray(self, section: str, subsection: str, **kwargs) -> np.ndarray:
        """Read data from *section*/*subsection* of the main KF file and return as NumPy array.

        All additional provided keyword arguments will be passed onto the numpy.array_ function.

        .. _numpy.array: https://docs.scipy.org/doc/numpy/reference/generated/numpy.array.html
        """
        return np.array(self.readkf(section, subsection), **kwargs)



class SCMJob(SingleJob):
    """Abstract class gathering common mechanisms for jobs with ADF Suite programs."""
    _result_type = SCMResults
    _top = ['title','units','define']
    _command = ''
    _subblock_end = 'subend'
    _legacy = ['band', 'dftb', 'uff']


    def __init__(self, **kwargs):
        SingleJob.__init__(self, **kwargs)
        if self.__class__._command in self.__class__._legacy:
            log("LEGACY WARNING: Job {} uses executable '{}' which is not present in AMS2018. Please use AMSJob (unless you're running an older version of ADFSuite)".format(self.name, self.__class__._command), 1)


    def get_input(self):
        """Generate the input file. This method is just a wrapper around :meth:`_serialize_input`.

        Each instance of |SCMJob| or |SCMResults| present as a value in ``settings.input`` branch is replaced with an absolute path to the main KF file of that job.
        """
        special = {
            SCMJob: lambda x: x.results._kfpath(),
            SCMResults: lambda x: x._kfpath(),
            KFFile: lambda x: x.path
        }
        return self._serialize_input(special)


    def get_runscript(self):
        """Generate a runscript. Returned string is of the form::

            $AMSBIN/name [-n nproc] <jobname.in [>jobname.out]

        ``name`` is taken from the class attribute ``_command``. ``-n`` flag is added if ``settings.runscript.nproc`` exists. ``[>jobname.out]`` is used based on ``settings.runscript.stdout_redirect``.
        """
        s = self.settings.runscript
        ret = '$AMSBIN/'+self._command
        if 'nproc' in s:
            ret += ' -n ' + str(s.nproc)
        ret += ' <"{}"'.format(self._filename('inp'))
        if s.stdout_redirect:
            ret += ' >"{}"'.format(self._filename('out'))
        ret += '\n\n'
        return AMSJob._slurm_env(self.settings) + ret


    def check(self):
        """Check if ``termination status`` variable from ``General`` section of main KF file equals ``NORMAL TERMINATION``."""
        try:
            status = self.results.readkf('General', 'termination status')
        except:
            return False
        if 'NORMAL TERMINATION' in status:
            if 'errors' in status:
                log('Job {} reported errors. Please check the the output'.format(self.name), 1)
                return False
            if 'warnings' in status:
                log('Job {} reported warnings. Please check the the output'.format(self.name), 1)
            return True
        return False


    def hash_input(self):
        """Calculate the hash of the input file.

        All instances of |SCMJob| or |SCMResults| present as values in ``settings.input`` branch are replaced with hashes of corresponding job's inputs.
        """
        special = {
            SCMJob: lambda x: x.hash_input(),
            SCMResults: lambda x: x.job.hash_input(),
            KFFile: lambda x: x.path
        }
        return sha256(self._serialize_input(special))


    def _serialize_input(self, special):
        """Transform all contents of ``setting.input`` branch into string with blocks, keys and values.

        On the highest level alphabetic order of iteration is modified: keys occuring in attribute ``_top`` are printed first. Special values can be indicated with *special* argument, which should be a dictionary having types of objects as keys and functions translating these types to strings as values.

        Automatic handling of ``molecule`` can be disabled with ``settings.ignore_molecule = True``.
        """

        def unspec(value):
            """Check if *value* is one of a special types and convert it to string if it is."""
            for spec_type in special:
                if isinstance(value, spec_type):
                    return special[spec_type](value)
            return value

        def serialize(key, value, indent):
            """Given a *key* and its corresponding *value* from the |Settings| instance produce a snippet of the input file representing this pair.

            If the value is a nested |Settings| instance, use recursive calls to build the snippet for the entire block. Indent the result with *indent* spaces.
            """
            ret = ''
            if isinstance(value, Settings):
                ret += ' '*indent + key
                if '_h' in value:
                    ret += ' ' + unspec(value['_h'])
                ret += '\n'

                i = 1
                while ('_'+str(i)) in value:
                    ret += serialize('', value['_'+str(i)], indent+2)
                    i += 1

                for el in value:
                    if not el.startswith('_'):
                        ret += serialize(el, value[el], indent+2)

                if indent == 0:
                    ret += 'end\n'
                else:
                    ret += ' '*indent + self._subblock_end + '\n'
            elif isinstance(value, list):
                for el in value:
                    ret += serialize(key, el, indent)
            elif value is '' or value is True:
                ret += ' '*indent + key + '\n'
            elif value is False:
                pass
            else:
                value = str(unspec(value))
                ret += ' '*indent + key
                if key != '' and not value.startswith('='):
                    ret += ' '
                ret += value + '\n'
            return ret


        use_molecule = ('ignore_molecule' not in self.settings) or (self.settings.ignore_molecule == False)
        if use_molecule:
            self._serialize_mol()

        inp = ''
        for item in self._top:
            item = self.settings.input.find_case(item)
            if item in self.settings.input:
                inp += serialize(item, self.settings.input[item], 0) + '\n'
        for item in self.settings.input:
            if item.lower() not in self._top:
                inp += serialize(item, self.settings.input[item], 0) + '\n'

        if use_molecule:
            self._remove_mol()
        return inp



    def _serialize_mol(self):
        """Process |Molecule| instance stored in ``molecule`` attribute and add it as relevant entries of ``settings.input`` branch. Abstract method."""
        raise PlamsError('Trying to run an abstract method SCMJob._serialize_mol()')


    def _remove_mol(self):
        """Remove from ``settings.input`` all entries added by :meth:`_serialize_mol`. Abstract method."""
        raise PlamsError('Trying to run an abstract method SCMJob._remove_mol()')


    @staticmethod
    def _atom_symbol(atom):
        """Return the atomic symbol of *atom*. Ensure proper formatting for ADFSuite input taking into account ``ghost`` and ``name`` entries in ``properties`` of *atom*."""
        smb = atom.symbol if atom.atnum > 0 else ''  #Dummy atom should have '' instead of 'Xx'
        if 'ghost' in atom.properties and atom.properties.ghost:
            smb = ('Gh.'+smb).rstrip('.')
        if 'name' in atom.properties:
            smb = (smb+'.'+str(atom.properties.name)).lstrip('.')
        return smb


    @classmethod
    def from_inputfile(cls, filename: str, heredoc_delimit: str = 'eor', **kwargs) -> 'SCMJob':
        """Construct a :class:`SCMJob` instance from an ADF inputfile.

        If a runscript is provide than this method will attempt to extract the input file based
        on the heredoc delimiter (see *heredoc_delimit*).

        """
        try:
            from scm.input_parser import InputParser
        except ImportError:  # Try to load the parser from $AMSHOME/scripting
            with UpdateSysPath():
                from scm.input_parser import InputParser

        s = Settings()
        with open(filename, 'r') as f:
            inp_file = parse_heredoc(f.read(), heredoc_delimit)

        with InputParser as parser:
            s.input = parser.to_settings(cls._command, inp_file)
        if not s.input:
            raise JobError(f"from_inputfile: failed to parse '{filename}'")

        # Extract a molecule from the input settings
        mol = cls.settings_to_mol(s)

        # Create and return the Job instance
        if mol is not None:
            return cls(molecule=mol, settings=s, **kwargs)
        else:
            s.ignore_molecule = True
            return cls(settings=s, **kwargs)


    @staticmethod
    def settings_to_mol(s: Settings) -> None:
        """An abstract method for extracting molecules from input settings (see :meth:`SCMJob.from_inputfile`)."""
        return None



import numpy as np
import os
import shutil

from os.path import join as opj

from ...core.basejob import SingleJob
from ...core.errors import FileError
from ...core.functions import log
from ...core.settings import Settings
from ...tools.units import Units
from .scmjob import SCMResults
from .scmjob import SCMJob, SCMResults


__all__ = ['ReaxFFJob', 'ReaxFFResults', 'load_reaxff_control', 'reaxff_control_to_settings']


class ReaxFFResults(SCMResults):
    _rename_map = {'reaxout.kf':'$JN.kf'}
    _kfext = '.kf'

    def _int2inp(self):
        return list(range(1, 1+len(self.job.molecule)))


class ReaxFFJob(SingleJob):
    _filenames = {'inp':'control', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err'}
    _result_type = ReaxFFResults
    ffield_path = opj('$AMSHOME','atomicdata','ForceFields','ReaxFF')
    default_cell_size = 100.0

    check = SCMJob.check


    def get_input(self):
        """Produce the ``control`` file based on key-value pairs present in ``settings.input.control`` branch."""

        s = self.settings.input.control
        ret = ''
        order = s._order if '_order' in s else iter(s)
        for key in order:
            ret += '{:>7} {:6}\n'.format(s[key], key)
        return ret


    def get_runscript(self):
        """Generate a runscript.

        Returned string is just ``$AMSBIN/reaxff``, possibly prefixed with ``export NSCM=(number)`` if ``settings.runscript.nproc`` is present.
        """
        s = self.settings.runscript
        ret = ''
        if 'nproc' in s:
            ret += 'export NSCM={}\n'.format(s.nproc)
        ret += '$AMSBIN/reaxff'
        if s.stdout_redirect:
            ret += ' >'+self._filename('out')
        ret += '\n\n'
        return ret


    def hash_input(self):
        """Disable hashing for ReaxFF jobs.

        It is a common task in molecular dynamics to run several trajectories with the same initial conditions. In such a case |RPM| would prevent second and all consecutive executions. Hence we decided to disable |RPM| for ReaxFF.

        If you wish to bring it back, simply put ``ReaxFFJob.hash_input = SingleJob.hash_inputs`` somehwere at the beginning of your script.
        """
        return None


    def _get_ready(self):
        """Prepare contents of the job folder for execution.

        Use the parent method from |SingleJob| to produce the runscript and the input file (``control``). Then create ``ffield`` and ``geo`` files using, respectively, :meth:`_write_ffield` and :meth:`_write_geofile`.

        Then copy to the job folder all files listed in ``settings.input.external``. The value of this key should either be a list of strings with paths to files or a dictionary (also |Settings|) with paths to files as values and names under which these files should be copied to the job folder as keys.
        """
        SingleJob._get_ready(self)
        self._write_ffield(self.settings.input.ffield)
        self._write_geofile(molecule=self.molecule, filename=opj(self.path, 'geo'), settings=self.settings.input.geo, description=self.name, lattice=True)

        if 'external' in self.settings.input:
            ext = self.settings.input.external
            if isinstance(ext, list):
                for val in ext:
                    if os.path.isfile(val):
                        shutil.copy(val, self.path)
                    else:
                        raise FileError('File {} not present'.format(val))
            elif isinstance(ext, dict):
                for key, val in ext.items():
                    if os.path.isfile(val):
                        shutil.copy(val, opj(self.path, key))
                    else:
                        raise FileError('File {} not present'.format(val))



    def _write_ffield(self, ffield):
        """Copy to the job folder a force field file indicated by *ffield**.

        *ffield* should be a string with a path to some external file or with a filename present in ``$AMSHOME/atomicdata/ForceFields/ReaxFF``. The location of this search folder is defined by ``ffield_path`` class attribute).

        Given file is always coied to the job folder as ``ffield``, due to ReaxFF program requirements.
        """

        if os.path.isfile(ffield):
            shutil.copy(ffield, opj(self.path, 'ffield'))
        else:
            path = os.path.expandvars(opj(self.ffield_path, ffield))
            if os.path.isfile(path):
                shutil.copy(path, opj(self.path, 'ffield'))
            else:
                raise FileError('settings.input.ffield={} is neither a path to a file nor an existing force field from {}'.format(ffield, self.ffield_path))



    def _write_geofile(self, molecule, filename, settings, description, lattice=False):
        """Write to *filename* a geo-file describing *molecule*.

        *settings* should be a |Settings| instance containing all the additional key-value pairs that should be present in the resulting geo-file. To obtain multiple occurrences of the same key in the geo-file, put all the values as a list in *settings*.

        *description* is the default value for ``DESCRP`` key. It is used only if ``descrp`` key is not present in *settings*.

        If *lattice* is ``True``, the information about periodicity is printed to the resulting geo-file with ``CRYSTX`` key. If the supplied *molecule* does not contain lattice vectors (or contains less then 3 of them), this method will add them (and hence alter *molecule*!).  The length of added vectors is defined by ``default_cell_size`` class attribute.

        *settings* can also be a single string with a path to a file -- in that case this file is copied as *filename* and all the rest of this method is skipped.

        .. note::

            If *lattice* is ``True`` and the lattice present in *molecule* does not follow ReaxFF convention (the third vector aligned with Z axis, the second one with YZ plane), this method will rotate the *molecule* to fulfill these requirements.
        """
        if isinstance(settings, str) and os.path.isfile(settings):
            shutil.copy(settings, opj(self.path, filename))
        else:
            header = ['BIOGRF 200\n']

            descrp = settings.find_case('descrp')
            if descrp not in settings:
                settings.descrp = description

            for key in settings:
                val = settings[key]
                if isinstance(val, list):
                    for el in val:
                        header.append('{:6} {}\n'.format(key.upper(), el))
                elif isinstance(val, tuple):
                    header.append(('{:6} '+'{} '*len(val)+'\n').format(key.upper(), *val))
                else:
                    header.append('{:6} {}\n'.format(key.upper(), val))

            if lattice is True:
                if molecule.align_lattice(convention='reax'):
                    log("The lattice supplied for job {} did not follow the convention required by ReaxFF. I rotated the whole system for you. You're welcome".format(self._full_name()), 3)

                f = lambda x: tuple([self.default_cell_size * int(i==x) for i in range(3)])
                while len(molecule.lattice) < 3:
                    molecule.lattice.append(f(len(molecule.lattice)))

                header.append('CRYSTX  {:10.5f} {:10.5f} {:10.5f} {:10.5f} {:10.5f} {:10.5f}\n'.format(*self._convert_lattice(molecule.lattice)))

            atoms = []
            for i,at in enumerate(molecule):
                newline = 'HETATM {:>5d} {:<2}               {: >10.5f}{: >10.5f}{: >10.5f} {:<2}     1 1  0.0\n'.format(i+1, at.symbol, *at.coords, at.symbol)
                atoms.append(newline)
            atoms.append('END\n')

            with open(opj(self.path, filename), 'w') as f:
                f.writelines(header)
                f.writelines(atoms)


    @staticmethod
    def _convert_lattice(lattice):
        """Convert a *lattice* expressed as three 3-dimensional vectors to (*a*, *b*, *c*, *alpha*, *beta*, *gamma*) format. Lengths of lattice vectors are expressed as *a*, *b* and *c*, angles between them as *alpha*, *beta*, *gamma*.
        """
        a, b, c = map(np.linalg.norm, lattice)
        al = np.dot(lattice[1], lattice[2])/(b*c)
        be = np.dot(lattice[0], lattice[2])/(a*c)
        ga = np.dot(lattice[0], lattice[1])/(a*b)
        al,be,ga = map(lambda x: Units.convert(np.arccos(x),'rad','deg'), [al,be,ga])
        return a, b, c, al, be, ga


#===========================================================================


def load_reaxff_control(filename, keep_order=True):
    """Return a |Settings| instance containing all data from an existing ``control`` file, indicated by *filename*.

    If *keep_order* is ``True``, the returned |Settings| instance is enriched with the ``_order`` key containing a list of all keys in the same order they were present in the loaded ``contol`` file.
    """

    if not os.path.isfile(filename):
        raise FileError('File {} not present'.format(filename))

    with open(filename, 'r') as f:
        lines = f.readlines()

    ret = Settings()
    if keep_order:
        ret._order = []

    for line in lines:
        if line.lstrip().startswith('#'):
            continue
        tmp = line.split()
        if len(tmp) > 1:
            value, key = tmp[:2]
            if len(key) > 6:
                log("load_reaxff_control: I am confused by this line:\n{}\nAre you sure {} is correct?".format(line, key), 3)
            if not value.replace('.','', 1).isdigit():
                 log("load_reaxff_control: I am confused by this line:\n{}\nAre you sure {} is correct?".format(line, value), 3)
            try:
                value = float(value) if '.' in value else int(value)
            except:
                pass
            ret[key] = value
            if keep_order:
               ret._order.append(key)
    return ret


def reaxff_control_to_settings(fpath:str) -> Settings:
    with open(fpath) as f:
        lines = [i.strip().split() for i in f]
    d = {}
    for i in lines:
        if len(i) > 1:
            try: v = int(i[0])
            except:
                try:    v = float(i[0])
                except: continue
            d[i[1]] = v

    s = Settings()
    if 'icheck' in d and d['icheck'] > 0:
        s.input.ams.task = 'singlepoint'
        if d['icheck'] == 1:
            s.input.ams.properties.gradients = True
    elif 'imetho' in d:
        v = d['imetho']
        if v == 0 or v == 2: # TODO: 2 is MD minimization, how should this be handled?
            s.input.ams.task = 'moleculardynamics'
        elif v == 1:
            s.input.ams.task = 'geometryoptimization'
        else:
            raise Exception(f"Could not understand imetho = {v}")
    else:
        raise Exception(f"Could not find a task. Is this a control file?")


    if s.input.ams.task == 'moleculardynamics':
        if 'nmdit'  in d: s.input.ams.moleculardynamics.nsteps   = d['nmdit']
        if 'tstep'  in d: s.input.ams.moleculardynamics.timestep = d['tstep']
        if 'mdtemp' in d: s.input.ams.moleculardynamics.initialvelocities.temperature = d['mdtemp']
        if 'iout2'  in d: s.input.ams.moleculardynamics.samplingfreq = d['iout2']
        if 'imdmet' in d:
            if d['imdmet'] == 1 or d['imdmet'] == 4: # NVT or NPT
                # T:
                s.input.ams.moleculardynamics.thermostat.type = 'Berendsen'
                s.input.ams.moleculardynamics.thermostat.temperature = d['mdtemp']
                if 'tdamp1' in d:
                    s.input.ams.moleculardynamics.thermostat.tau = d['tdamp1']
                if 'itdmet' in d:
                    s.input.ams.moleculardynamics.thermostat.berendsenapply = 'Local' if d['itdmet'] == 0 else 'Global'
                # P:
                if d['imdmet'] == 4: # NPT
                    s.input.ams.moleculardynamics.barostat = 'Berendsen'
                    if 'mdpres' in d:
                        s.input.ams.moleculardynamics.barostat.pressure = d['mdpres']*1e9
                    if 'pdamp1' in d:
                        s.input.ams.moleculardynamics.barostat.tau = d['pdamp1']

    elif s.input.ams.task == 'geometryoptimization':
        grads_set, fire = False, True
        if 'endmm' in d:
            s.input.ams.geometryoptimization.convergence.gradients = d['endmm'] * Units.conversion_ratio('kcal/mol','hartree')
            grads_set = True
        if 'imaxmo':
            v = d['imaxmo']
            s.input.ams.geometryoptimization.pretendconverged = True
            if v == 0:
                s.input.ams.geometryoptimization.method = 'ConjugateGradients'
                fire = False
            else:
                s.input.ams.geometryoptimization.convergence.step = v*1e-6
        if 'imaxit' in d:
            s.input.ams.geometryoptimization.pretendconverged = True
            s.input.ams.geometryoptimization.maxiterations = d['imaxit']
        if 'icelop' in d and d['icelop'] > 0:
            s.input.ams.geometryoptimization.optimizelattice = True
            fire = True # CG does not support lattice optimization

        if grads_set and fire:
            # FIRE calculates rms on atomic gradients (1/Natoms) whereas others calculate components (1/3Natoms)
            s.input.ams.geometryoptimization.convergence.gradients *= np.sqrt(3)
            s.input.ams.geometryoptimization.method = 'FIRE'

    return s



import os

from os.path import join as opj

from ...core.basejob import SingleJob
from ...core.private import saferun
from ...core.results import Results
from ...core.settings import Settings

__all__ = ['DiracJob', 'DiracResults']



class DiracResults(Results):
    """A class for result of computation done with DIRAC."""
    _rename_map = {'DFCOEF':'$JN.dfcoef', 'GRIDOUT':'$JN.grid', 'dirac.xml':'$JN.xml',}



    def collect(self):
        """After collecting the files produced by job execution with parent method :meth:`Results.collect<scm.plams.core.results.Results.collect>` append the ``pam`` output to the regular output file.
        """
        Results.collect(self)
        pamfile = self.job._filename('out')
        process = saferun(['grep', 'output file', pamfile], cwd=self.job.path)
        output = process.stdout.decode()
        diracfile = output.split(':')[-1].strip()
        if diracfile in self.files:
            pampath = opj(self.job.path, pamfile)
            diracpath = opj(self.job.path, diracfile)
            with open(pampath, 'r') as f:
                pamoutput = f.readlines()
            with open(diracpath, 'r') as f:
                diracoutput = f.readlines()
            with open(pampath, 'w') as f:
                f.writelines(diracoutput)
                f.write('\n\n   '+'*'*74+'\n')
                f.write('   '+'*'*30+'  pam output  '+'*'*30+'\n')
                f.write('   '+'*'*74+'\n\n')
                f.writelines(pamoutput)
            os.remove(diracpath)
        self.refresh()



class DiracJob(SingleJob):
    """A class representing a single computational job with DIRAC."""

    _result_type = DiracResults
    _top = ['dirac']
    _filenames = {'inp':'$JN.inp', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err'}


    def __init__(self, **kwargs):
        SingleJob.__init__(self, **kwargs)
        self.settings.runscript.pam.noarch = True
        self.settings.runscript.pam.get = ['DFCOEF', 'GRIDOUT', 'dirac.xml']



    def _get_ready(self):
        """Before generating runscript and input with parent method :meth:`SingleJob._get_ready<scm.plams.core.basejob.SingleJob._get_ready>` add proper ``mol`` and ``inp`` entries to ``self.settings.runscript.pam``. If already present there, ``mol`` will not be added.
        """
        s = self.settings.runscript.pam
        if 'mol' not in s:
            s.mol = self.name+'.xyz'
            with open(opj(self.path, self.name+'.xyz'), 'w') as f:
                f.write(str(len(self.molecule)) + '\n\n')
                for atom in self.molecule:
                    suffix = 'b={block}' if hasattr(atom,'block') else ''
                    f.write(atom.str(suffix=suffix)+'\n')
        s.inp = self._filename('inp')
        SingleJob._get_ready(self)



    def get_input(self):
        """Transform all contents of ``input`` branch of ``settings`` into string with blocks, subblocks, keys and values.

        On the highest level alphabetic order of iteration is modified: keys occuring in class attribute ``_top`` are printed first. See :ref:`dirac-input` for details.
        """
        is_empty = lambda x: isinstance(x, Settings) and len(x) == 0

        def parse_key(key, value):
            ret = '.' + key.upper() + '\n'
            if not (value is True or is_empty(value)):
                if isinstance(value, list):
                    for i in value:
                        ret += str(i) + '\n'
                else:
                    ret += str(value) + '\n'
            return ret

        def parse_block(block):
            enabler = '_en'
            ret = '**' + block.upper() + '\n'
            s = self.settings.input[block]
            for k,v in s.items():
                if not isinstance(v, Settings) or is_empty(v):
                    ret += parse_key(k, v)
            for k,v in s.items():
                if isinstance(v, Settings) and enabler in v:
                    ret += parse_key(k, v[enabler])
            for k,v in s.items():
                if isinstance(v, Settings) and len(v) > 0:
                    ret += '*' + k.upper() + '\n'
                    for kk,vv in v.items():
                        if kk != enabler:
                            ret += parse_key(kk, vv)
            return ret

        inp = ''
        for block in self._top:
            if block in self.settings.input:
                inp += parse_block(block)
        for block in self.settings.input:
            if block not in self._top:
                inp += parse_block(block)
        inp += '*END OF INPUT\n'
        return inp



    def get_runscript(self):
        """Generate a runscript. Returned string is a ``pam`` call followed by option flags generated based on ``self.settings.runscript.pam`` contents. See :ref:`dirac-runscript` for details."""
        r = self.settings.runscript.pam
        ret = 'pam'
        for k,v in r.items():
            ret += ' --%s'%k
            if v is not True:
                if isinstance(v, list):
                    ret += '="%s"' % ' '.join(v)
                else:
                    ret += '='+str(v)

        if self.settings.runscript.stdout_redirect:
            ret += ' >'+self._filename('out')
        ret += '\n\n'
        return ret



    def check(self):
        """Check if the calculation was successful by examining the last line of ``pam`` output."""
        s = self.results.grep_output('exit           :')[0]
        status = s.split(':')[-1].strip()
        return status == 'normal'



#!/usr/bin/env python

import numpy
from ..tools.periodic_table import PT
from ..mol.molecule import Molecule
from ..mol.atom import Atom
from ..core.settings import Settings
from .xyzfile import XYZTrajectoryFile
from .xyzfile import data_from_xyzcomment

__all__ = ['XYZHistoryFile']

class XYZHistoryFile (XYZTrajectoryFile) :
        """
        Class representing an XYZ file containing a molecular simulation history with varying numbers of atoms

        An instance of this class has the following attributes:

        *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual XYZ file
        *   ``position``    -- The frame to which the cursor is currently pointing in the XYZ file
        *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
        *   ``elements``    -- The elements of the atoms in the system at the current frame

        An |XYZHistoryFile| object behaves very similar to a regular file object.
        It has read and write methods (:meth:`read_next` and :meth:`write_next`) 
        that read and write from/to the position of the cursor in the ``file_object`` attribute. 
        If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
        the cursor to any frame in the file and reads from there.
        The amount of information stored in memory is kept to a minimum, as only information from the current frame
        is ever stored.

        Reading and writing to and from the files can be done as follows::

            >>> from scm.plams import XYZHistoryFile

            >>> xyz = XYZHistoryFile('old.xyz')
            >>> mol = xyz.get_plamsmol()

            >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

            >>> for i in range(xyz.get_length()) :
            >>>     crd,cell = xyz.read_frame(i,molecule=mol)
            >>>     xyzout.write_next(molecule=mol)

        The above script reads information from the XYZ file ``old.xyz`` into the |Molecule| object ``mol``
        in a step-by-step manner.
        The |Molecule| object is then passed to the :meth:`write_next` method of the new |XYZHistoryFile|
        object corresponding to the new xyz file ``new.xyz``.

        The exact same result can also be achieved by iterating over the instance as a callable

            >>> xyz = XYZHistoryFile('old.xyz')
            >>> mol = xyz.get_plamsmol()

            >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

            >>> for crd,cell in xyz(mol) :
            >>>     xyzout.write_next(molecule=mol)

        This procedure requires all coordinate information to be passed to and from the |Molecule| object
        for each frame, which can be time-consuming.
        It is therefore also possible to bypass the |Molecule| object when reading through the frames::

            >>> xyz = XYZHistoryFile('old.xyz')

            >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

            >>> for crd,cell in xyz :
            >>>     xyzout.write_next(coords=crd,elements=xyz.elements)

        By default the write mode will create a minimal version of the XYZ file, containing only elements
        and coordinates. 
        Additional information can be written to the file by supplying additional arguments
        to the :meth:`write_next` method. 
        The additional keywords `step` and `energy` trigger the writing of a remark containing
        the molecule name, the step number, the energy, and the lattice vectors.

            >>> mol = Molecule('singleframe.xyz')

            >>> xyzout = XYZHistoryFile('new.xyz',mode='w')
            >>> xyzout.set_name('MyMol')

            >>> xyzout.write_next(molecule=mol, step=0, energy=5.)
        """

        def __init__ (self, filename, mode='r', fileobject=None, ntap=None) :
                """
                Initiates an XYZHistoryFile object

                * ``filename``   -- The path to the XYZ file
                * ``mode``       -- The mode in which to open the XYZ file ('r' or 'w')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms can be passed here
                """
                XYZTrajectoryFile.__init__(self,filename,mode,fileobject,ntap)

                self.input_elements = self.elements[:]

        def _is_endoffile (self) :
                """
                If the end of file is reached, return coords and cell as None
                """
                end = False
                line = self.file_object.readline()
                if len(line) == 0 :
                        end = True
                        return end
                nats = int(line.split()[0])
                for i in range(nats+1) :
                        line = self.file_object.readline()
                        if len(line) == 0 :
                                end = True
                                break
                return end

        def _read_coordinates (self, molecule) :
                """
                Read the coordinates at current step
                """
                # Find the number of atoms
                line = self.file_object.readline()
                if len(line) == 0 :
                        return None, None     # End of file is reached
                nats = int(line.split()[0])
                line = self.file_object.readline()

                # Handle the comment line
                cell = None
                historydata = data_from_xyzcomment(line)
                if 'Lattice' in historydata :
                        cell = historydata['Lattice']
                        del historydata['Lattice']
                if self.include_historydata :
                        self.historydata = historydata

                # Read coordinates and elements
                coords = []
                elements = []
                for i in range(nats) :
                        line = self.file_object.readline() 
                        words = line.split()
                        coords.append([float(w) for w in words[1:4]])
                        elements.append(words[0])

                # If the elements changed, update the molecule
                if elements != self.elements :
                        self.elements = elements
                        self.coords = numpy.array(coords)
                        # Rebuild the molecule (bonds will disappear for now)
                        if isinstance(molecule,Molecule) :
                                for at in reversed(molecule.atoms) :
                                        molecule.delete_atom(at)
                                molecule.properties = Settings()
                                for el in elements :
                                        atom = Atom(PT.get_atomic_number(el))
                                        molecule.add_atom(atom)
                else :
                        self.coords[:] = coords

                # Assign the data to the molecule object
                if isinstance(molecule,Molecule) :
                        self._set_plamsmol(self.coords,cell,molecule,bonds=None)

                return coords, cell

        def write_next (self,coords=None,molecule=None,elements=None,cell=[0.,0.,0.],conect=None,historydata=None) :
                """ 
                Write frame to next position in trajectory file

                * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
                * ``molecule`` -- A molecule object to read the molecular data from
                * ``elements`` -- The element symbols of the atoms in the system
                * ``cell``     -- A set of lattice vectors or cell diameters
                * ``energy``   -- An energy value to be written to the remark line
                * ``conect``   -- A dictionary containing connectivity info (not used)
                * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

                The ``historydata`` dictionary can contain for example:
                ('Step','Energy'), the frame number and the energy respectively

                .. note::

                        Either ``coords`` and ``elements`` or ``molecule`` are mandatory arguments
                """
                if isinstance(molecule,Molecule) :
                        coords, cell, elements = self._read_plamsmol(molecule)[:3]
                self.elements = elements
                cell = self._convert_cell(cell)
                        
                self._write_moldata(coords, cell, historydata)
                
                self.position += 1



import collections
import math
import numpy as np

from ..core.errors import UnitsError


__all__ = ['Units']


class Units:
    """A singleton class for unit converter.

    All values are based on `2014 CODATA recommended values <http://physics.nist.gov/cuu/Constants>`_.

    The following constants and units are supported:

    *   constants:

        -   ``speed_of_light`` (also ``c``)
        -   ``electron_charge`` (also ``e``)
        -   ``Avogadro_constant`` (also ``NA``)
        -   ``Bohr_radius``

    *   distance:

        -   ``Angstrom``, ``A``
        -   ``Bohr``, ``au``, ``a.u.``
        -   ``nm``
        -   ``pm``

    *   reciprocal distance:

        -   ``1/Angstrom``, ``1/A``, ``Angstrom^-1``, ``A^-1``,
        -   ``1/Bohr``, ``Bohr^-1``

    *   angle:

        -    ``degree``, ``deg``,
        -    ``radian``, ``rad``,
        -    ``grad``
        -    ``circle``

    *   energy:

        -   ``au``, ``a.u.``, ``Hartree``
        -   ``eV``
        -   ``kcal/mol``
        -   ``kJ/mol``
        -   ``cm^-1``, ``cm-1``

    *   dipole moment:

        -   ``au``, ``a.u.``
        -   ``Cm``
        -   ``Debye``, ``D``

    *   forces:

        -   All energy units divided by angstrom or bohr, for example 
        -   ``eV/angstrom``
        -   ``hartree/bohr``

    *   hessian:

        -   All energy units divided by angstrom^2 or bohr^2, for example 
        -   ``eV/angstrom^2``
        -   ``hartree/bohr^2``

    *   pressure:

        -   All energy units divided by angstrom^3 or bohr^3, for example 
        -   ``eV/angstrom^3``
        -   ``hartree/bohr^3``
        -   And some more:
        -   ``Pa``
        -   ``GPa``
        -   ``bar``
        -   ``atm``


    Example::

        >>> print(Units.constants['speed_of_light'])
        299792458
        >>> print(Units.constants['e'])
        1.6021766208e-19
        >>> print(Units.convert(123, 'angstrom', 'bohr'))
        232.436313431
        >>> print(Units.convert([23.32, 145.0, -34.7], 'kJ/mol', 'kcal/mol'))
        [5.573613766730401, 34.655831739961755, -8.293499043977056]
        >>> print(Units.conversion_ratio('kcal/mol', 'kJ/mol'))
        4.184


    """

    constants = {}
    constants['Bohr_radius']                         =  0.529177210903  #http://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
    constants['Avogadro_constant'] = constants['NA'] =  6.022140857e23   #http://physics.nist.gov/cgi-bin/cuu/Value?na
    constants['speed_of_light'] = constants['c']     =  299792458   #http://physics.nist.gov/cgi-bin/cuu/Value?c
    constants['electron_charge'] = constants['e']    =  1.6021766208e-19   #http://physics.nist.gov/cgi-bin/cuu/Value?e
    constants['Boltzmann'] = constants['k_B'] = 1.380649e-23 #J/K


    distance = {}
    distance['A'] = distance['Angstrom']                 =  1.0
    distance['Bohr'] = distance['a.u.'] = distance['au'] =  1.0 / constants['Bohr_radius']
    distance['nm']                                       = distance['A'] / 10.0
    distance['pm']                                       = distance['A'] * 100.0
    distance['m']                                        = distance['A'] * 1e-10

    rec_distance = {}
    rec_distance['1/A'] = rec_distance['1/Ang'] = rec_distance['1/Angstrom'] = rec_distance['A^-1'] = rec_distance['Ang^-1'] = rec_distance['Angstrom^-1'] = 1.0
    rec_distance['1/m'] = rec_distance['m^-1'] = 1e10
    rec_distance['1/Bohr'] = rec_distance['Bohr^-1'] = constants['Bohr_radius']

    energy = {}
    energy['au'] = energy['a.u.'] = energy['Hartree'] = energy['Ha'] =  1.0
    energy['eV']                                                     =  27.211386245988   #http://physics.nist.gov/cgi-bin/cuu/Value?hrev
    energy['kJ/mol']                                                 =  4.359744650e-21 * constants['NA']  #http://physics.nist.gov/cgi-bin/cuu/Value?hrj
    energy['J']                                                      =  4.359744650e-18
    energy['kcal/mol']                                               =  energy['kJ/mol'] / 4.184
    energy['cm^-1'] = energy['cm-1']                                 =  219474.6313702   #http://physics.nist.gov/cgi-bin/cuu/Value?hrminv
    energy['K'] = energy['J'] / constants['k_B']

    mass = {}
    mass['au'] = mass['a.u.'] = mass['amu'] = 1.0
    mass['kg'] = 1.66053906660e-27
    mass['g'] = mass['kg'] * 1e3

    angle = {}
    angle['degree'] =  angle['deg'] = 1.0
    angle['radian'] =  angle['rad'] = math.pi / 180.0
    angle['grad']   =  100.0 / 90.0
    angle['circle'] =  1.0 / 360.0

    dipole = {}
    dipole['au'] = dipole['a.u.'] =  1.0
    dipole['Cm']                  =  constants['e'] * constants['Bohr_radius'] * 1e-10
    dipole['Debye'] = dipole['D'] =  dipole['Cm'] * constants['c']* 1e21

    forces = {}
    hessian = {}
    stress = {}
    for k,v in energy.items():
        forces[k+'/Angstrom'] = forces[k+'/Ang'] = forces[k+'/A'] = v * rec_distance['1/Angstrom']
        hessian[k+'/Angstrom^2'] = hessian[k+'/Ang^2'] = hessian[k+'/A^2'] = v  * rec_distance['1/Angstrom']**2
        stress[k+'/Angstrom^3'] = stress[k+'/Ang^3'] = stress[k+'/A^3'] = v  * rec_distance['1/Angstrom']**3
        forces[k+'/bohr'] = forces[k+'/au'] = forces[k+'/a.u.'] = v * rec_distance['1/Bohr']
        hessian[k+'/bohr^2'] = hessian[k+'/au^2'] = hessian[k+'/a.u.^2'] = v * rec_distance['1/Bohr']**2
        stress[k+'/bohr^3'] = stress[k+'/au^3'] = stress[k+'/a.u.^3'] = v * rec_distance['1/Bohr']**3
        forces[k+'/m'] = v * rec_distance['1/m']
        hessian[k+'/m^2'] = v  * rec_distance['1/m']**2
        stress[k+'/m^3'] = v  * rec_distance['1/m']**3
    forces['au'] = forces['a.u.'] = forces['Ha/bohr']
    hessian['au'] = hessian['a.u.'] = hessian['Ha/bohr^2']
    stress['au'] = stress['a.u.'] = stress['Ha/bohr^3']
    stress['Pa'] = stress['J/m^3']
    stress['GPa'] = stress['Pa'] * 1e-9
    stress['bar'] = stress['Pa'] * 1e-5
    stress['atm'] = stress['bar'] / 1.01325



    dicts = {}
    dicts['distance'] = distance
    dicts['energy'] = energy
    dicts['mass'] = mass
    dicts['angle'] = angle
    dicts['dipole'] = dipole
    dicts['reciprocal distance'] = rec_distance
    dicts['forces'] = forces
    dicts['hessian'] = hessian
    dicts['stress'] = stress


    def __init__(self):
        raise UnitsError('Instances of Units cannot be created')


    @classmethod
    def find_unit(cls, unit):
        ret = {}
        for quantity in cls.dicts:
            for k in cls.dicts[quantity]:
                if k.lower() == unit.lower():
                    ret[quantity] = k
        return ret


    @classmethod
    def conversion_ratio(cls, inp, out):
        """Return conversion ratio from unit *inp* to *out*."""
        if inp == out:
            return 1.
        inps = cls.find_unit(inp)
        outs = cls.find_unit(out)
        common = set(inps.keys()) & set(outs.keys())
        if len(common) > 0:
            quantity = common.pop()
            d = cls.dicts[quantity]
            return d[outs[quantity]]/d[inps[quantity]]
        else:
            if len(inps) == 0 and len(outs) == 0:
                raise UnitsError("Unsupported units: '{}' and '{}'".format(inp, out))
            if len(inps) > 0 and len(outs) > 0:
                raise UnitsError("Invalid unit conversion: '{}' is a unit of {} and '{}' is a unit of {}".format(inp, ', '.join(list(inps.keys())), out, ', '.join(list(outs.keys()))))
            else: #exactly one of (inps,outs) empty
                invalid, nonempty = (out,inps) if len(inps) else (inp,outs)
                if len(nonempty) == 1:
                    quantity = list(nonempty.keys())[0]
                    raise UnitsError("Invalid unit conversion: {} is not supported. Supported units for {}: {}".format(invalid, quantity, ', '.join(list(cls.dicts[quantity].keys()))))
                else:
                    raise UnitsError("Invalid unit conversion: {} is not a supported unit for {}".format(invalid, ', '.join(list(nonempty.keys()))))




    @classmethod
    def convert(cls, value, inp, out):
        """Convert *value* from unit *inp* to *out*.

        *value* can be a single number or a container (list, tuple, numpy.array etc.). In the latter case a container of the same type and length is returned. Conversion happens recursively, so this method can be used to convert, for example, a list of lists of numbers, or any other hierarchical container structure. Conversion is applied on all levels, to all values that are numbers (also numpy number types). All other values (strings, bools etc.) remain unchanged.
        """
        if inp == out:
            return value
        if value is None or isinstance(value, (bool, str)):
            return value
        if isinstance(value, collections.abc.Iterable):
            t = type(value)
            if t == np.ndarray:
                t = np.array
            v = [cls.convert(i, inp, out) for i in value]
            return t(v)
        if isinstance(value, (int, float, np.generic)):
            return value * cls.conversion_ratio(inp,out)
        return value



from ...core.errors import PlamsError

__all__ = ["AMSPipeError", "AMSPipeDecodeError", "AMSPipeLogicError", "AMSPipeRuntimeError", "AMSPipeUnknownVersionError", "AMSPipeUnknownMethodError", "AMSPipeUnknownArgumentError", "AMSPipeInvalidArgumentError"]

class AMSPipeError(PlamsError):
    """Base class for exceptions mapped to errors defined by the pipe protocol.

    Not to be used directly.
    """
    def __init__(self, message, argument=None):
        self.status = _class2code[self.__class__]
        self.message = message
        self.method = None
        self.argument = argument

    def __str__(self):
        return ": ".join(s for s in (self.method, self.argument, self.message) if s is not None)

    @staticmethod
    def from_message(msg):
        exc = _code2class[msg["status"]](msg.get("message"), msg.get("argument"))
        exc.method = msg.get("method")

        return exc

    def to_message(self):
        msg = {"status": self.status}
        if self.message is not None:
            msg["message"] = self.message
        if self.method is not None:
            msg["method"] = self.method
        if self.argument is not None:
            msg["argument"] = self.argument

        return msg



class AMSPipeDecodeError(AMSPipeError):
    pass

class AMSPipeLogicError(AMSPipeError):
    pass

class AMSPipeRuntimeError(AMSPipeError):
    pass

class AMSPipeUnknownVersionError(AMSPipeError):
    pass

class AMSPipeUnknownMethodError(AMSPipeError):
    pass

class AMSPipeUnknownArgumentError(AMSPipeError):
    pass

class AMSPipeInvalidArgumentError(AMSPipeError):
    pass


_code2class = {
    1: AMSPipeDecodeError,
    2: AMSPipeLogicError,
    3: AMSPipeRuntimeError,
    4: AMSPipeUnknownVersionError,
    5: AMSPipeUnknownMethodError,
    6: AMSPipeUnknownArgumentError,
    7: AMSPipeInvalidArgumentError
}
_class2code = {c: i for i, c in _code2class.items()}



from ..core.basejob import MultiJob
from ..core.results import Results
from ..core.settings import Settings
from ..mol.molecule import Molecule
from ..interfaces.adfsuite.ams import AMSJob


__all__ = ['ADFFragmentJob', 'ADFFragmentResults']


class ADFFragmentResults(Results):

    def get_properties(self):
        return self.job.full.results.get_properties()

    def get_main_molecule(self):
        return self.job.full.results.get_main_molecule()

    def get_input_molecule(self):
        return self.job.full.results.get_input_molecule()

    def get_energy(self, unit='au'):
        return self.job.full.results.get_energy(unit)

    def get_dipole_vector(self, unit='au'):
        return self.job.full.results.get_dipole_vector(unit)

    def get_energy_decomposition(self):
        energy_section = self.job.full.results.read_rkf_section('Energy', file='adf')
        ret = {}
        for k in ['Electrostatic Energy', 'Kinetic Energy', 'Elstat Interaction', 'XC Energy']:
            ret[k] = energy_section[k]
        return ret


class ADFFragmentJob(MultiJob):
    _result_type = ADFFragmentResults

    def __init__(self, fragment1=None, fragment2=None, full_settings=None, **kwargs):
        MultiJob.__init__(self, **kwargs)
        self.fragment1 = fragment1.copy() if isinstance(fragment1, Molecule) else fragment1
        self.fragment2 = fragment2.copy() if isinstance(fragment2, Molecule) else fragment2
        self.full_settings = full_settings or Settings()

    def prerun(self):
        self.f1 = AMSJob(name='frag1', molecule=self.fragment1, settings=self.settings)
        self.f2 = AMSJob(name='frag2', molecule=self.fragment2, settings=self.settings)

        for at in self.fragment1:
            at.properties.suffix = 'adf.f=subsystem1'
        for at in self.fragment2:
            at.properties.suffix = 'adf.f=subsystem2'

        self.full = AMSJob(name = 'full',
            molecule = self.fragment1 + self.fragment2,
            settings = self.settings + self.full_settings)

        self.full.settings.input.adf.fragments.subsystem1 = (self.f1, 'adf')
        self.full.settings.input.adf.fragments.subsystem2 = (self.f2, 'adf')

        self.children = [self.f1, self.f2, self.full]





import numpy as np

from ..core.basejob import MultiJob
from ..core.errors import JobError
from ..core.results import Results
from ..interfaces.adfsuite.adf import ADFJob
from ..tools.units import Units

__all__ = ['CSHessianADFJob', 'CSHessianADFResults']

try:
    from scm.csh import CSHessian
except ImportError:
    __all__ = []


class CSHessianADFResults(Results):
    def get_hessian(self):
        return self.hessians[-1]

    def get_all_hessians(self):
        return self.hessians


class CSHessianADFJob(MultiJob):
    _result_type = CSHessianADFResults

    def __init__(self, molecule, basistype, displacement=0.01, start=0.2, step=1, conv_threshold=1e-2, **kwargs):
        MultiJob.__init__(self, children={}, **kwargs)
        self.molecule = molecule
        self.basistype = basistype
        self.disp = displacement
        self.start = start
        self.threshold = conv_threshold
        self.step = step


    def prerun(self):
        self.restartjob = ADFJob(name='restart', molecule=self.molecule, settings=self.settings.gradient)
        self.restartjob.parent = self
        self.restartjob.run()

        self.N = 3 * len(self.molecule)
        self.perm = np.random.permutation(self.N)
        self.pick = max(int(self.N * self.start), 1)
        self.batch = self.perm[:self.pick]

        self.basis = self.get_basis(self.basistype)
        self.csh = CSHessian(self.basis)
        self.samp = self.csh.get_sampling_basis()

        self.scaling = [None] * self.N
        self.results.hessians = []
        for i in self.batch:
            for step in [-1,1]:
                self.children[(i, step)] = self.create_forces_job(i, step)


    def create_forces_job(self, column, step):
        vec = self.samp[:, column]
        vec.shape = (-1,3)

        #scale vec in such a way that most displaced atom is moved by self.disp
        max_displ = 0.0
        for v in vec:
            max_displ = max(max_displ, np.linalg.norm(v))
        self.scaling[column] = self.disp / max_displ

        mol = self.molecule.copy()
        for at, v in zip(mol, vec):
            at.translate(self.scaling[column] * step * v)

        newjob = ADFJob(name=f'{column}_{step}', molecule=mol, settings=self.settings.gradient)
        newjob.settings.input.restart.file = self.restartjob
        newjob.settings.input.restart.nogeo = 'True'
        return newjob


    def get_basis(self, arg):
        if isinstance(arg, str):
            if arg in self.settings.basis:
                s = self.settings.basis[arg]
                self.basisjob = s.jobtype(name='basis', molecule=self.molecule, settings=s)
                self.basisjob.parent = self
                self.basisjob.run()
                hess = s.get_hessian(self.basisjob.results)
                hess.shape = (self.N, self.N)

            elif arg == 'cart':
                return np.eye(self.N)

            else:
                raise JobError(f"CSHessianADFJob: I don't understand 'basistype={arg}' argument. My Settings should contain .basis.{arg} branch, but they do not.")

        else:
            try:
                arg = np.array(arg)
                arg.shape = (self.N, self.N)
            except:
                raise JobError(f"CSHessianADFJob: argument supplied as 'basistype' cannot be transformed into numpy array of size ({self.N},{self.N})")
            return arg

        evals, evecs = np.linalg.eigh(hess)
        return evecs


    def compare(self, x, y):
        return np.linalg.norm(x-y) < self.threshold


    def new_batch(self):
        previous = self.pick
        self.pick = min(self.pick + self.step, self.N)
        return self.perm[previous:self.pick]


    def new_children(self):
        for i in self.batch:
            v1 = np.array(self.children[(i,-1)].results.readkf('GeoOpt', 'Gradients_InputOrder'))
            v2 = np.array(self.children[(i,1)].results.readkf('GeoOpt', 'Gradients_InputOrder'))
            grad_cart = Units.convert(v2 - v1, 'bohr', 'angstrom') / (2*self.scaling[i])
            grad_samp = np.dot(self.samp.T, grad_cart)
            self.csh.add_column(i, grad_samp)

        new_hessian = self.csh.estimate()
        if self.results.hessians and self.compare(self.results.hessians[-1], new_hessian):
            return {}

        self.results.hessians.append(new_hessian)
        self.batch = self.new_batch()
        ret = {}
        for i in self.batch:
            for step in [-1,1]:
                ret[(i,step)] = self.create_forces_job(i, step)
        return ret





# Workflow with PLAMS

# A couple of useful function for extracting results:

@add_to_class(AMSResults)
def get_excitations(results):
    """Returns excitation energies (in eV) and oscillator strenghts (in Debye)."""
    if results.job.ok():
        exci_energies_au = results.readrkf('Excitations SS A','excenergies',file='engine')
        oscillator_str_au = results.readrkf('Excitations SS A','oscillator strengths',file='engine')
        # The results are stored in atomic units. Convert them to more convenient units:
        exci_energies = Units.convert(exci_energies_au,'au','eV')
        oscillator_str = Units.convert(oscillator_str_au,'au','Debye')
        return exci_energies, oscillator_str
    else:
        return [],[]


@add_to_class(AMSResults)
def has_good_excitations(results, min_energy, max_energy, oscillator_str_threshold=1E-4):
    """Returns True if there is at least one excitation with non-vanishing oscillator strenght
       in the energy range [min_energy, max_energy]. Unit for min_energy and max energy: eV. """
    exci_energies, oscillator_str = results.get_excitations()
    for e,o in zip(exci_energies, oscillator_str):
        if e > min_energy and e < max_energy and o > oscillator_str_threshold:
            return True
    return False


# Calculation settings:
# =====================

# Settings for geometry optimization with the AMS driver:
go_sett = Settings()
go_sett.input.ams.Task = 'GeometryOptimization'
go_sett.input.ams.GeometryOptimization.Convergence.Gradients = 1.0e-4

# Settings for single point calculation with the AMS driver
sp_sett = Settings()
sp_sett.input.ams.Task = 'SinglePoint'

# Settings for the DFTB engine (including excitations)
dftb_sett = Settings()
dftb_sett.input.dftb.Model = 'SCC-DFTB'
dftb_sett.input.dftb.ResourcesDir = 'QUASINANO2015'
dftb_sett.input.dftb.Properties.Excitations.TDDFTB.calc = 'singlet'
dftb_sett.input.dftb.Properties.Excitations.TDDFTB.lowest = 10
dftb_sett.input.dftb.Occupation.Temperature = 5.0

# Settings for the geometry optimization with the ADF engine
adf_sett = Settings()
adf_sett.input.adf.Basis.Type = 'DZP'
adf_sett.input.adf.NumericalQuality = 'Basic'

# Settings for the excitation calculation using the ADF engine
adf_exci_sett = Settings()
adf_exci_sett.input.adf.Basis.Type = 'TZP'
adf_exci_sett.input.adf.XC.GGA = 'PBE'
adf_exci_sett.input.adf.NumericalQuality = 'Basic'
adf_exci_sett.input.adf.Symmetry = 'NoSym'
adf_exci_sett.input.adf.Excitations.lowest = 10
adf_exci_sett.input.adf.Excitations.OnlySing = ''


# Import all xyz files in the folder 'molecules'

molecules = read_molecules('molecules')


print("Step 1: prescreening with DFTB")
print("==============================")

promising_molecules = {}

for name, mol in molecules.items():
  dftb_job = AMSJob(name='DFTB_'+name, molecule=mol, settings=go_sett+dftb_sett)
  dftb_job.run()

  if dftb_job.results.has_good_excitations(1,6):
    promising_molecules[name] = dftb_job.results.get_main_molecule()

print(f"Found {len(promising_molecules)} promising molecules with DFTB")


print("Step 2: Optimization and excitations calculation with ADF")
print("=========================================================")

for name, mol in promising_molecules.items():
  adf_go_job = AMSJob(name='ADF_GO_'+name, molecule=mol, settings=go_sett+adf_sett)
  adf_go_job.run()

  optimized_mol = adf_go_job.results.get_main_molecule()

  adf_exci_job = AMSJob(name='ADF_exci_'+name, molecule=optimized_mol, 
                        settings=sp_sett+adf_exci_sett)
  adf_exci_job.run()

  if adf_exci_job.results.has_good_excitations(2,4):
    print(f"Molecule {name} has excitation(s) satysfying our criteria!")
    print(optimized_mol)
    exci_energies, oscillator_str = adf_exci_job.results.get_excitations()
    print("Excitation energy [eV], oscillator strength:")
    for e,o in zip(exci_energies, oscillator_str):
        print(f"{e:8.4f}, {o:8.4f}")



class GammaResults(Results):

    @staticmethod
    def get_difference(job, jobplus):
        """Calculate the difference between HOMO and IP.
        *jobplus* should be the counterpart of *job* with one less electron."""
        homo = job.results.readrkf('Properties','HOMO', file='engine')
        IP = jobplus.results.get_energy() - job.results.get_energy()
        return IP + homo

    def get_J(self):
        N = GammaResults.get_difference(self.job.children[1], self.job.children[2])
        A = GammaResults.get_difference(self.job.children[0], self.job.children[1])
        return (N**2 + A**2)**0.5


class GammaJob(MultiJob):
    _result_type = GammaResults

    def __init__(self, molecule, gamma, charge, spins, **kwargs):
        MultiJob.__init__(self, **kwargs)
        self.molecule = molecule
        self.charge = charge
        self.spins = spins
        self.gamma = gamma

    def prerun(self):
        charges = [self.charge-1, self.charge, self.charge+1]
        for charge, spin in zip(charges, self.spins):
            name = '{}_charge_{}'.format(self.name, charge)
            newjob = AMSJob(name=name, molecule=self.molecule, settings=self.settings)
            newjob.settings.input.ams.System.charge = charge
            newjob.settings.input.adf.xc.rangesep = "gamma={:f}".format(self.gamma)
            if spin != 0:
                newjob.settings.input.adf.unrestricted = True
                newjob.settings.input.adf.SpinPolarization = spin

            self.children.append(newjob)


def gamma_scan(gammas, settings, molecule, name='scan', charge=0, spins=(1,0,1)):
    """Calculate values of J function for given range of gammas.

    Arguments:
    gammas   - list of gamma values to calculate the J function for
    settings - Settings object for an ADF calculation
    molecule - Molecule object with the system of interest
    name     - base name of all the jobs
    charge   - base charge of the system of interest. The J function is going to be
               calculated based on two systems: with charge, and charge-1
    spins    - values of spin polarization for jobs with, respectively, charge-1, 
               charge and charge +1

    In other words, if charge=X and spins=(a,b,c) the three resulting jobs
    are going to have the following values for charge and spin:

    Charge=X-1  SpinPolarization=a
    Charge=X    SpinPolarization=b
    Charge=X+1  SpinPolarization=c

    Returns a list of pairs (gamma, J) of the same length as the parameter *gammas*
    """
    jobs = [GammaJob(molecule=molecule, settings=settings, gamma=g,
            charge=charge, spins=spins, name=name+'_gamma_'+str(g)) for g in gammas]
    results = [j.run() for j in jobs]
    js = [r.get_J() for r in results]
    return list(zip(gammas, js))


# =============================================================
# Now we simply use the gamma_scan function to find the optimal 
# gamma value for a toy system (H2)
# =============================================================


import numpy as np
import multiprocessing

# Run as many jobs in parallel as there are cores:
config.default_jobrunner = JobRunner(parallel=True, maxjobs=multiprocessing.cpu_count())

# Settings of the ADF calculations
# ================================

s = Settings()
s.input.ams.task = 'SinglePoint'
s.input.adf.basis.type = 'DZP'
s.input.adf.basis.core = 'None'
s.input.adf.xc.gga = 'PBE'
s.input.adf.xc.xcfun = True
s.runscript.nproc = 1

# The molecule (here we just use H2)
# ==================================

mol = Molecule()
mol.add_atom(Atom(symbol='H', coords=(0,0,-0.3540)))
mol.add_atom(Atom(symbol='H', coords=(0,0, 0.3540)))

# The list of gamma values
# ========================

# Here we scan just a few values for gamma. 
# In practice, you want to scan a wider range and smaller step.
gammas = np.around(np.arange(1.2, 1.9, 0.2), decimals=3)

results = gamma_scan(gammas, s, mol)

print('== Results ==')
print('gamma \t J')
for g,j in results:
    print('{:.4f} \t {:.8f}'.format(g,j))
print('Optimal gamma value: {:.4f}'.format(min(results,key=lambda x:x[1])[0]))



#!/usr/bin/env python

import os
import numpy
from..mol.molecule import Bond
from ..core.errors import PlamsError

__all__ = ['TrajectoryFile']

class TrajectoryFile (object) :
        """
        Abstract class that represents a generic trajectory file
        """
        def __init__ (self, filename, mode='r', fileobject=None, ntap=None) :
                """
                Would create a generic trajectory  file object

                * ``filename``   -- The path to the DCD file
                * ``mode``       -- The mode in which to open the DCD file ('rb' or 'wb')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
                """
                self.position = 0
                if filename is not None :
                        fileobject = open(filename,mode)
                self.file_object = fileobject
                if self.file_object is not None :
                        self.mode = self.file_object.mode

                self.ntap = 0
                if ntap is not None :
                        self.ntap = ntap
                self.firsttime = True
                self.coords = numpy.zeros((self.ntap,3))                # Only for reading purposes, to avoid creating the array each time

                # PLAMS molecule related settings
                self.elements = ['H']*self.ntap
                self.current_molecule = None
                self.store_molecule = True # Even if True, the molecule attribute is only stored during iteration

        def __iter__ (self) :
                """
                Magic method that makes this an iterator

                Note: Iterates starting from current cursor position
                """
                if self.mode[0] == 'r' :
                        if self.current_molecule is None and self.store_molecule :
                                self.current_molecule = self.get_plamsmol()
                        for i in range(len(self)-self.position) :
                                crd,cell = self.read_next(molecule=self.current_molecule)
                                yield crd,cell
                        self.current_molecule = None
                else :
                        raise PlamsError('Cannot iterate over writable trajectoryfile')

        @property
        def molecule (self) :
                """
                Returns the current molecule, which exists only when the object is used as an iterator
                """
                if not self.store_molecule :
                        raise PlamsError ('Try setting the store_molecule attribute to True')
                elif self.current_molecule is None :
                        raise PlamsError ('molecule attribute only exists during iteration')
                return self.current_molecule

        def __call__ (self, molecule=None, read=True) :
                """
                Magic method that makes an instance of this class into a callable
                """
                if self.mode[0] == 'r' :
                        for i in range(len(self)-self.position) :
                                crd,cell = self.read_next(molecule, read)
                                yield crd,cell
                else :
                        raise PlamsError('Cannot iterate over writable trajectoryfile')

        def __len__ (self) :
                """
                Magic method that allows checking the length of an iterator
                """
                return self.get_length()

        def __enter__ (self) :
                """
                Magic method that allows use with *with* statement
                """
                return self

        def __exit__ (self, exc_type, exc_val, exc_tb) :
                """
                Magic method that allows use with *with* statement
                """
                self.close()

        def __del__ (self) :
                """
                If self.file_object is a regular fileobject, close it
                """
                try :
                        self.file.close()
                except AttributeError :
                        pass

        def get_elements (self) :
                """
                Get the elements attribute
                """
                return self.elements

        def set_elements (self, elements) :
                """
                Sets the elements attribute (needed in write mode).

                *   ``elements`` -- A list containing the element symbol of each atom
                """
                self.elements = elements

        def _read_header (self) :
                """
                Set up info required for reading frames

                Sets self.ntap and self.coords to proper value/size
                """
                pass

        def get_plamsmol (self) :
                """
                Extracts a PLAMS molecule object from the XYZ trajectory file
                """
                from scm.plams import Molecule
                oldposition = self.position
                self.rewind()
                coords, cell = self.read_next()
                plamsmol = Molecule.from_elements(self.elements)
                plamsmol.from_array(coords)

                # Return to original position
                self.rewind()
                for i in range(oldposition) :
                        self.read_next(read=False)

                return plamsmol

        def close (self) :
                """
                Cleanly close and garbage collect the file
                """
                del(self)

        def read_next (self, molecule=None, read=True) :
                """
                Reads the relevant info from current frame
                """
                if not read and not self.firsttime :
                        return self._move_cursor_without_reading()

                # Read the coordinates and cell info from self.file_object
                self.coords = None
                self.cell = None

                # Finalize
                if self.firsttime :
                        self.firsttime = False

                # Place the values into the provided molecule object
                if isinstance(molecule,Molecule) :
                        self._set_plamsmol(self.coords,cell,molecule)

                self.position += 1
                return self.coords, cell

        def _move_cursor_without_reading (self) :
                """
                Move the cursor forward and return coords as empty array
                """
                cell = numpy.zeros((3,3))
                if self._is_endoffile() :
                        return None, None
                self.position += 1
                coords = numpy.array([])
                return coords, cell

        def _is_endoffile (self) :
                """
                Reads and checks If the end of file is reached.
                """
                pass

        def read_frame (self, frame, molecule=None) :
                """
                Reads the relevant info from frame ``frame`` and returns it, or stores it in ``molecule``

                * ``frame``    -- The frame number to be read from the file
                * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
                """
                if frame < self.position :
                        nframes = abs(frame - self.position)
                        self.rewind(nframes)
                steps = frame - self.position
                
                for i in range(steps) :
                        crd,cell = self.read_next(read=False)
                        if crd is None :
                                break
                
                crd,cell = self.read_next(molecule)
                if crd is None :
                        print('Not enough frames!')
                
                return crd,cell

        def _set_plamsmol (self, coords, cell, plamsmol, bonds=None) :
                """
                If molecule objects have been passed as arguments, update their coordinates and lattice
                """
                plamsmol.from_array(coords)
                if cell is not None :
                        if cell[0][0]**2 + cell[0][1]**2 + cell[0][2]**2 > 0. :
                                plamsmol.lattice = cell
                                if isinstance(cell,numpy.ndarray) :
                                        plamsmol.lattice = cell.tolist()
                if bonds is not None :
                        plamsmol.delete_all_bonds()
                        for bond in bonds :
                                b = Bond(plamsmol[bond[0]],plamsmol[bond[1]])
                                plamsmol.add_bond(b)

        def write_next (self, coords=None, molecule=None, cell=[0.,0.,0.], energy=0., step=None, conect=None) :
                """
                Write the coordinates to the next available spot in the file
                """
                if isinstance(molecule,Molecule) :
                        coords, cell = self._read_plamsmol(molecule)[:2]
                cell = self._convert_cell(cell)
                # Write to self.file_object
                self.position += 1

        def _convert_cell (self, cell) :
                """
                Check format of cell, and convert to vectors if needed
                """
                if cell is None :
                        return cell
                # The cell can be passed as a list of three values (assumed orthorhombic)
                if len(cell)==3 and isinstance(cell[0],float) or isinstance(cell[0],numpy.float64) :
                        cell = numpy.diag(cell)
                else :
                        cell = numpy.array(cell)
                # For non-periodic systems there will be three cell vectors of length 0.
                if cell[0][0]**2 + cell[0][1]**2 + cell[0][2]**2 == 0. :
                        cell = None
                return cell

        def _read_plamsmol (self, plamsmol) :
                """
                Read the coordinates and cell vectors from the molecule objects, if provided

                Note: For non-periodic systems the cell will be read as [0.,0.,0.]
                """
                coords = plamsmol.as_array()
                cell = [0.,0.,0.]
                if len(plamsmol.lattice) > 0 : cell = plamsmol.lattice
                elements = [at.symbol for at in plamsmol.atoms]
                # Get the connection table
                plamsmol.set_atoms_id()
                conect = {}
                for iat,atom in enumerate(plamsmol.atoms) :
                        neighbors = plamsmol.neighbors(atom)
                        if len(neighbors) > 0 :
                                conect[iat+1] = [neighbor.id for neighbor in neighbors]
                # Get the properties
                props = [at.properties.suffix if 'suffix' in at.properties else '' for at in plamsmol.atoms]
                if sum([len(s) for s in props]) == 0 : props = None
                return coords, cell, elements, conect, props

        def rewind (self,nframes=None) :
                """ 
                Rewind the file either by ``nframes`` or to the first frame

                *   ``nframes`` -- The number of frames to rewind
                """
                if nframes is None or nframes == self.position :
                        # Go back to the beginning
                        self._rewind_to_first_frame()
                elif nframes > self.position :
                        raise PlamsError('Trying to rewind too much!')
                else :  
                        # Go back nframes geometries
                        self._rewind_n_frames(nframes)

        def _rewind_to_first_frame(self) :
                """
                Rewind the file to the first frame
                """
                self.file_object.seek(0)
                self.firsttime = True
                self.position = 0
                self._read_header()
                
        def _rewind_n_frames(self,nframes) :
                """
                Rewind the file by nframes frames
                """
                pass

        def get_length (self) :
                """
                Get the number of frames in the file
                """
                oldposition = self.position
                while 1 :
                        crd,cell = self.read_next(read=False)
                        if crd is None :
                                break
                #size = self.position-1
                size = self.position

                # Go back to the original position
                self.rewind()
                for i in range(oldposition) :
                        crd,cell = self.read_next(read=False)

                return size

        def read_last_frame (self, molecule=None) :
                """
                Reads the last frame from the file
                """
                step = 0
                while 1 :
                        crd,cell = self.read_next(read=False)
                        if crd is None :
                                break
                        step += 1

                # If the current position was already the end of the file, this fails, so a failsafe is built in:                
                if step == 0 :
                        step = self.position

                self.rewind(1)
                crd,cell = self.read_next(molecule=molecule)
                return crd,cell



import copy
import functools
import glob
import inspect
import operator
import os
import shutil
import threading

from os.path import join as opj
from subprocess import PIPE

from .private import saferun
from .errors import ResultsError, FileError
from .functions import config, log


__all__ = ['Results']



def _caller_name_and_arg(frame):
    """Extract information about name and arguments of a function call from a *frame* object"""
    if frame is None:
        return None, None
    caller_name = frame.f_code.co_name
    caller_varnames = frame.f_code.co_varnames
    caller_arg = None
    if len(caller_varnames) > 0:
        try:
            loc = frame.f_locals
        except:
            loc = {}
        if caller_varnames[0] in loc:
            caller_arg = loc[caller_varnames[0]]
    return caller_name, caller_arg


def _privileged_access():
    """Analyze contents of the current stack to find out if privileged access to the |Results| methods should be granted.

    Privileged access is granted to two |Job| methods: |postrun| and :meth:`~scm.plams.core.basejob.Job.check`, but only if they are called from :meth:`~scm.plams.core.basejob.Job._finalize` of the same |Job| instance.
    """
    from .basejob import Job
    for frame in inspect.getouterframes(inspect.currentframe()):
        cal, arg = _caller_name_and_arg(frame[0])
        prev_cal, prev_arg = _caller_name_and_arg(frame[0].f_back)
        if cal in ['postrun', 'check'] and prev_cal == '_finalize' and arg == prev_arg and isinstance(arg, Job):
                return True
    return False


def _restrict(func):
    """Decorator that wraps methods of |Results| instances.

    Whenever decorated method is called, the status of associated job is checked. Depending of its value access to the method is granted, refused or the calling thread is forced to wait for the right :ref:`event<event-objects>` to be set.
    """

    @functools.wraps(func)
    def guardian(self, *args, **kwargs):
        if not self.job:
            raise ResultsError('Using Results not associated with any Job')

        if self.job.status in ['successful', 'copied']:
            return func(self, *args, **kwargs)

        elif self.job.status in ['preview']:
            if config.ignore_failure:
                log('WARNING: Trying to obtain results of job {} run in a preview mode. Returned value is None'.format(self.job.name), 3)
                return None
            else:
                raise ResultsError('Using Results associated with job run in a preview mode')

        elif self.job.status in ['deleted']:
            raise ResultsError('Using Results associated with deleted job')

        elif self.job.status in ['crashed', 'failed']:
            if func.__name__ == 'wait': #waiting for crashed of failed job should not trigger any warnings/exceptions
                cal, arg = _caller_name_and_arg(inspect.currentframe())
                if isinstance(arg, Results):
                    return func(self, *args, **kwargs)
            if config.ignore_failure:
                log('WARNING: Trying to obtain results of crashed or failed job {}'.format(self.job.name), 3)
                try:
                    ret = func(self, *args, **kwargs)
                except:
                    log('Obtaining results of {} failed. Returned value is None'.format(self.job.name), 3)
                    return None
                log('Obtaining results of {} successful. However, no guarantee that they make sense'.format(self.job.name), 3)
                return ret
            else:
                raise ResultsError('Using Results associated with crashed or failed job')

        elif self.job.status in ['created', 'started', 'registered', 'running']:
            log('Waiting for job {} to finish'.format(self.job.name), 1)
            if _privileged_access():
                self.finished.wait()
            else:
                self.done.wait()
            return func(self, *args, **kwargs)

        elif self.job.status in ['finished']:
            if _privileged_access():
                return func(self, *args, **kwargs)
            log('Waiting for job {} to finish'.format(self.job.name), 1)
            self.done.wait()
            return func(self, *args, **kwargs)

    return guardian



#===========================================================================
#===========================================================================
#===========================================================================



class _MetaResults(type):
    """Metaclass for |Results|. During new |Results| instance creation it wraps all methods with :func:`_restrict` decorator ensuring proper synchronization and thread safety. Methods listed in ``_dont_restrict`` as well as "magic methods" are not wrapped."""
    _dont_restrict = ['refresh', 'collect', '_clean']
    def __new__(meta, name, bases, dct):
        for attr in dct:
            if not (attr.endswith('__') and attr.startswith('__')) and callable(dct[attr]) and (attr not in _MetaResults._dont_restrict):
                dct[attr] = _restrict(dct[attr])
        return type.__new__(meta, name, bases, dct)



#===========================================================================
#===========================================================================
#===========================================================================



class Results(metaclass=_MetaResults):
    """General concrete class for job results.

    ``job`` attribute stores a reference to associated job. ``files`` attribute is a list with contents of the job folder. ``_rename_map`` is a class attribute with the dictionary storing the default renaming scheme.

    Bracket notation (``myresults[filename]``) can be used to obtain full absolute paths to files in the job folder.

    Instance methods are automatically wrapped with the "access guardian" that ensures thread safety (see |parallel|).
    """
    _rename_map = {}

    def __init__(self, job):
        self.job = job
        self.files = []
        self.finished = threading.Event()
        self.done = threading.Event()


    def refresh(self):
        """Refresh the contents of the ``files`` list. Traverse the job folder (and all its subfolders) and collect relative paths to all files found there, except files with ``.dill`` extension.

        This is a cheap and fast method that should be used every time there is a risk the contents of the job folder changed and ``files`` is no longer up-to-date. For proper working of various PLAMS elements it is crucial that ``files`` always contains up-to-date information about the contents of the job folder.

        All functions and methods defined in PLAMS that could change the state of the job folder refresh the ``files`` list, so there is no need to manually call :meth:`~Results.refresh` after, for example, :meth:`~Results.rename`. If you are implementing a new method of that kind, please don't forget about refreshing.
        """
        self.files = []
        for pth, dirs, files in os.walk(self.job.path):
            relpath = os.path.relpath(pth, self.job.path)
            self.files += [opj(relpath, x) if relpath != '.' else x for x in files]
        self.files = [x for x in self.files if not x.endswith('.dill')]


    def collect(self):
        """Collect the files present in the job folder after execution of the job is finished. This method is simply :meth:`~Results.refresh` followed by renaming according to the ``_rename_map``.

        If you wish to override this function, you have to call the parent version at the beginning.
        """
        self.refresh()
        for old, new in self.__class__._rename_map.items():
            old = old.replace('$JN', self.job.name)
            new = new.replace('$JN', self.job.name)
            if old in self.files:
                os.rename(opj(self.job.path, old), opj(self.job.path, new))
                self.files[self.files.index(old)] = new
        self.refresh()


    def wait(self):
        """wait()
        Wait for associated job to finish.

        .. technical::

            This is **not** an abstract method. It does exactly what it should: nothing. All the work is done by :func:`_restrict` decorator that is wrapped around it.
        """
        pass


    def grep_file(self, filename, pattern='', options=''):
        """grep_file(filename, pattern='', options='')
        Execute ``grep`` on a file given by *filename* and search for *pattern*.

        Additional ``grep`` flags can be passed with *options*, which should be a single string containing all flags, space separated.

        Returned value is a list of lines (strings). See ``man grep`` for details.
        """
        cmd = ['grep'] + [pattern] + options.split()
        return self._process_file(filename, cmd)


    def grep_output(self, pattern='', options=''):
        """grep_output(pattern='', options='')
        Shortcut for :meth:`~Results.grep_file` on the output file."""
        try:
            output = self.job._filename('out')
        except AttributeError:
            raise ResultsError('Job {} does not seem to be an instance of SingleJob, it does not have _filenames dictionary'.format(self.job.name))
        except KeyError:
            raise ResultsError('Job {} does not have an output'.format(self.job.name))
        return self.grep_file(output, pattern, options)


    def awk_file(self, filename, script='', progfile=None, **kwargs):
        """awk_file(filename, script='', progfile=None, **kwargs)
        Execute an AWK script on a file given by *filename*.

        The AWK script can be supplied in two ways: either by directly passing the contents of the script (should be a single string) as the *script* argument, or by providing the path (absolute or relative to *filename*) to a text file with an AWK script as the *progfile* argument. If *progfile* is not ``None``, *script* is ignored.

        Other keyword arguments (*\*\*kwargs*) can be used to pass additional variables to AWK (see ``-v`` flag in AWK manual)

        Returned value is a list of lines (strings). See ``man awk`` for details.
        """
        cmd = ['awk']
        for k,v in kwargs.items():
            cmd += ['-v', '{}={}'.format(k,v)]
        if progfile:
            if os.path.isfile(progfile):
                cmd += ['-f', progfile]
            else:
                raise FileError('File {} not present'.format(progfile))
        else:
            cmd += [script]
        return self._process_file(filename, cmd)


    def awk_output(self, script='', progfile=None, **kwargs):
        """awk_output(script='', progfile=None, **kwargs)
        Shortcut for :meth:`~Results.awk_file` on the output file."""
        try:
            output = self.job._filename('out')
        except AttributeError:
            raise ResultsError('Job {} does not seem to be an instance of SingleJob, it does not have _filenames dictionary'.format(self.job.name))
        except KeyError:
            raise ResultsError('Job {} does not have an output'.format(self.job.name))
        return self.awk_file(output, script, progfile, **kwargs)


    def rename(self, old, new):
        """rename(old, new)
        Rename a file from ``files``. In both *old* and *new* the shortcut ``$JN`` for job name can be used."""
        old = old.replace('$JN', self.job.name)
        new = new.replace('$JN', self.job.name)
        self.refresh()
        if old in self.files:
            os.rename(opj(self.job.path, old), opj(self.job.path, new))
            self.files[self.files.index(old)] = new
        else:
            raise FileError('File {} not present in {}'.format(old, self.job.path))


    def get_file_chunk(self, filename, begin=None, end=None, match=0, inc_begin=False, inc_end=False, process=None):
        """get_file_chunk(filename, begin=None, end=None, match=0, inc_begin=False, inc_end=False, process=None)

        Extract a chunk of a text file given by *filename*, consisting of all the lines between a line containing *begin* and a line containing *end*.

        *begin* and *end* should be simple strings (no regular expressions allowed) or ``None`` (in that case matching is done from the beginning or until the end of the file). If multiple blocks delimited by *begin* end *end* are present in the file, *match* can be used to indicate which one should be printed (*match*=0 prints all of them). *inc_begin* and *inc_end* can be used to include the delimiting lines in the final result (by default they are excluded).

        The returned value is a list of strings. *process* can be used to provide a function executed on each element of this list before returning it.
        """
        current_match = 0
        ret = []
        switch = (begin == None)

        append = lambda x: ret.append(x.rstrip('\n')) if (match in [0,current_match]) else None

        with open(self[filename], 'r') as f:
            for line in f:
                if switch and end and (end in line):
                    switch = False
                    if inc_end: append(line)
                    if match == current_match: break
                if switch:
                    append(line)
                if (not switch) and begin and (begin in line):
                    switch = True
                    current_match += 1
                    if inc_begin: append(line)

        return list(map(process, ret)) if process else ret


    def get_output_chunk(self, begin=None, end=None, match=0, inc_begin=False, inc_end=False, process=None):
        """get_output_chunk(begin=None, end=None, match=0, inc_begin=False, inc_end=False, process=None)
        Shortcut for :meth:`~Results.get_file_chunk` on the output file."""
        try:
            output = self.job._filename('out')
        except AttributeError:
            raise ResultsError('Job {} is not an instance of SingleJob, it does not have an output'.format(self.job.name))
        return self.get_file_chunk(output, begin, end, match, inc_begin, inc_end, process)


    def recreate_molecule(self):
        """Recreate the input molecule for the corresponding job based on files present in the job folder. This method is used by |load_external|.

        The definiton here serves as a deafult fall-back template preventing |load_external| from crashing when a particular |Results| subclass does not define it's own :meth:`recreate_molecule`.
        """
        return None


    def recreate_settings(self):
        """Recreate the input |Settings| instance for the corresponding job based on files present in the job folder. This method is used by |load_external|.

        The definiton here serves as a deafult fall-back template preventing |load_external| from crashing when a particular |Results| subclass does not define it's own :meth:`recreate_settings`.
        """
        return None



#=======================================================================


    def _clean(self, arg):
        """Clean the job folder. *arg* should be a string or a list of strings. See |cleaning| for details."""
        if arg == 'all':
            return

        path = self.job.path
        absfiles = [opj(path, f) for f in self.files]
        childnames = [child.name for child in self.job] if hasattr(self.job, 'children') else []
        if arg in ['none', [], None]:
            [os.remove(f) for f in absfiles if os.path.isfile(f)]

        elif isinstance(arg, list):
            rev = False
            if arg[0] == '-':
                rev = True
                arg = arg[1:]

            absarg = []
            for i in arg:
                s = i.replace('$JN', self.job.name)
                if s.find('$CH') != -1:
                    absarg += [opj(path, s.replace('$CH', ch)) for ch in childnames]
                else:
                    absarg.append(opj(path, s))

            if absarg:
                absarg = functools.reduce(operator.iadd, map(glob.glob, absarg))

            for f in absfiles:
                if (f in absarg) == rev and os.path.isfile(f):
                    os.remove(f)
                    log('Deleting file '+f, 5)

        else:
            log('WARNING: {} is not a valid keep/save argument'.format(arg), 3)
        self.refresh()


    def _copy_to(self, newresults):
        """_copy_to(newresults)
        Copy these results to *newresults*.

        This method is used when |RPM| discovers an attempt to run a job identical to a previously run job. Instead of the execution, results of the previous job are copied/linked to the new one.

        This method is called from |Results| of the old job and *newresults* should be |Results| of the new job. The goal is to faithfully recreate the state of this |Results| instance in ``newresults``. To achieve that, all the contents of the job folder are copied (or hardlinked, if your platform allows that and ``job.settings.link_files`` is ``True``) to other's job folder. Moreover, all attributes of this |Results| instance (other than ``job`` and ``files``) are exported to *newresults* using :meth:`~Results._export_attribute` method.
        """
        for name in self.files:
            newname = Results._replace_job_name(name, self.job.name, newresults.job.name)
            oldpath = opj(self.job.path, name)
            newpath = opj(newresults.job.path, newname)
            os.makedirs(os.path.dirname(newpath), exist_ok=True)
            if os.name == 'posix' and self.job.settings.link_files is True:
                os.link(oldpath, newpath)
            else:
                shutil.copy(oldpath, newpath)
            newresults.files.append(newname)
        for k,v in self.__dict__.items():
            if k in ['job', 'files', 'done', 'finished']: continue
            newresults.__dict__[k] = self._export_attribute(v, newresults)


    def _export_attribute(self, attr, other):
        """_export_attribute(attr, other)
        Export this instance's attribute to *other*. This method should be overridden in your |Results| subclass if it has some attributes that are not properly handled by :func:`python3:copy.deepcopy`.

        *other* is the |Results| instance, *attr* is the **value** of the attribute to be copied. See :meth:`SCMJob._export_attribute<scm.plams.interfaces.adfsuite.scmjob.SCMResults._export_attribute>` for an example implementation.
        """
        return copy.deepcopy(attr)


    @staticmethod
    def _replace_job_name(string, oldname, newname):
        """If *string* starts with *oldname*, maybe followed by some extension, replace *oldname* with *newname*."""
        return string.replace(oldname, newname) if (os.path.splitext(string)[0] == oldname) else string


#=======================================================================


    def __getitem__(self, name):
        """Magic method to enable bracket notation. Elements from ``files`` can be used to get absolute paths."""
        name = name.replace('$JN', self.job.name)
        if name in self.files:
            return opj(self.job.path, name)
        else:
            raise FileError('File {} not present in {}'.format(name, self.job.path))



    def _process_file(self, filename, command):
        """_process_file(filename, command)
        Skeleton for all file processing methods. Execute *command* (should be a list of strings) on *filename* and return output as a list of lines.
        """
        filename = filename.replace('$JN', self.job.name)
        if filename in self.files:
            process = saferun(command + [filename], cwd=self.job.path, stdout=PIPE)
            if process.returncode != 0:
                return []
            ret = process.stdout.decode().splitlines()
            return ret
        else:
            raise FileError('File {} not present in {}'.format(filename, self.job.path))





"""
Run DFTB+ with plams
contributed by Patrick Melix
based on code from Michal Handzlik

see documentation for an example
"""
from os.path import join as opj

from ...core.basejob import SingleJob
from ...core.settings import Settings
from ...core.results import Results
from ...tools.units import Units
from ...core.errors import MoleculeError
from ...mol.molecule import Molecule

__all__ = ['DFTBPlusJob', 'DFTBPlusResults']


class DFTBPlusResults(Results):
    """A Class for handling DFTB+ Results."""
    _outfile = 'detailed.out'
    _xyzout = 'geo_end.xyz'
    _genout = 'geo_end.gen'


    def get_molecule(self):
        """get_molecule()
        Return the molecule from the 'geo_end.gen' file. If there is an Error, try to read the 'geo_end.xyz' file.
        """
        try:
            #The .gen file contains the cell, ASE can read it
            mol = Molecule(self[self._genout], inputformat='ase')
        except MoleculeError:
            #Fallback if no ASE found
            mol = Molecule(filename=self[self._xyzout])
        except:
            mol = Molecule()
        return mol


    def get_energy(self, string='Total energy', unit='au'):
        """get_energy(string='Total energy', unit='au')
        Return the energy given in the output with the description *string*, expressed in *unit*. Defaults to ``Total energy`` and ``au``.
        """
        try:
            energy = float(self.grep_file(self._outfile, pattern=string+':')[0].split()[2])
            energy = Units.convert(energy, 'au', unit)
        except:
            energy = float('nan')
        return energy


    def get_atomic_charges(self):
        """get_atomic_charges()
        Returns a dictonary with atom numbers and their charges, ordering is the same as in the input.
        """
        try:
            atomic_charges = {}
            string = self.awk_file(self._outfile,script='/Net atomic charges/{do_print=1} NF==0 {do_print=0 }do_print==1 {print}')
            for line in string:
                if string[0] == line or string[1] == line:
                    continue
                l = line.split()
                atomic_charges[l[0]] = float(l[1])
        except:
            atomic_charges = {}
        return atomic_charges



class DFTBPlusJob(SingleJob):
    """A class representing a single computational job with DFTB+.
       Only supports molecular coordinates, no support for lattice yet."""
    _result_type = DFTBPlusResults
    _filenames = {'inp':'dftb_in.hsd', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err', 'gen': '$JN.gen'}


    def get_input(self):
      """Transform all contents of ``setting.input`` branch into string with blocks, keys and values.

      Automatic handling of ``molecule`` can be disabled with ``settings.ignore_molecule = True``.
      """
      def parse(key, value, indent=''):
          if value is True:
              value = ''

          ret = indent + key
          if key != '' and value != '':
              ret += ' ='

          if isinstance(value, Settings):
              if '_h' in value:
                  ret += ' ' + value['_h']
              ret += ' {\n'

              i = 1
              while ('_'+str(i)) in value:
                  ret += parse('', value['_'+str(i)], indent+'  ')
                  i += 1

              for el in value:
                  if not el.startswith('_'):
                      ret += parse(el, value[el], indent+'  ')
              ret += indent + '}'
          else:
              ret += ' ' + str(value)
          ret += '\n'
          return ret

      inp = ''
      use_molecule = ('ignore_molecule' not in self.settings) or (self.settings.ignore_molecule == False)
      if use_molecule:
          self._parsemol()

      for item in self.settings.input:
          inp += parse(item, self.settings.input[item])

      if use_molecule:
          self._removemol()
      return inp


    def _parsemol(self):
        #use ASE to write molecule if available
        if 'ase' in Molecule._writeformat:
            filename = opj(self.path, self._filename('gen'))
            self.molecule.write(filename, outputformat='ase', format='gen')
            self.settings.input.geometry._h = 'GenFormat'
            self.settings.input.geometry._1 = '<<< '+self._filename('gen')

        else:
            #Old way of handling gen-format ourselves, delete if ASE becomes obligatory
            atom_types = {}
            n = 1
            atoms_line = ''
            for atom in self.molecule:
                if atom.symbol not in atom_types:
                    atoms_line += atom.symbol + ' '
                    atom_types[atom.symbol] = n
                    n += 1

            #check PBC
            lattice = []
            geomType = 'C'
            for vec in self.molecule.lattice:
                if not all(isinstance(x, (int,float)) for x in vec):
                    raise ValueError("Non-Number in Lattice Vectors, not compatible with DFTBPlus")

                lattice.append(vec)
                geomType = 'S'

            self.settings.input.geometry._h = 'GenFormat'
            self.settings.input.geometry._1 = '%i %s'%(len(self.molecule),geomType)
            self.settings.input.geometry._2 = atoms_line
            self.settings.input.geometry._3 = ''

            for i,atom in enumerate(self.molecule):
                self.settings.input.geometry['_'+str(i+4)] = ('%5i'%(i+1)) + atom.str(symbol=str(atom_types[atom.symbol]))

            if len(vec) > 0:
                j = i + 1
                #origin
                self.settings.input.geometry['_'+str(j+4)] = '0.0 0.0 0.0'
                j += 1
                for i, vec in enumerate(lattice):
                    self.settings.input.geometry['_'+str(i+j+4)] = '%f %f %f'%(vec)


    def _removemol(self):
        if 'geometry' in self.settings.input:
            del self.settings.input.geometry


    def get_runscript(self):
        """dftb+ has to be in your $PATH!"""
        ret = 'dftb+ '
        if self.settings.runscript.stdout_redirect:
            ret += ' >' + self._filename('out')
        ret += '\n\n'
        return ret


    def check(self):
        """Returns true if 'ERROR!' is not found in the output."""
        s = self.results.grep_output('ERROR!')
        return len(s) == 0



from os.path import join as opj
import numpy as np

from ...core.basejob  import SingleJob
from ...core.settings import Settings
from ...core.errors import PlamsError
from ...mol.molecule  import Molecule

__all__ = ['CrystalJob','mol2CrystalConf']



class CrystalJob(SingleJob):
    """
    A class representing a single computational job with `CRYSTAL <https://www.crystal.unito.it/>`
    Use Crystal version >= 14, lower versions have an even stricter and more complicated input.
    """
    _command = 'crystal'
    _filenames = {'inp':'INPUT', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err'}

    def get_input(self):
        """
        Transform all contents of ``input`` branch of ``settings`` into string.
        """

        #these are the geometry keys from the manual, first mandatory block
        _geom_keys = ['CRYSTAL','SLAB','POLYMER','HELIX','MOLECULE','EXTERNAL','DLVINPUT']

        #second mandatory block: basis set
        #we use the option to end the geometry section with BASISSET instead of END,
        #because that is much more like PLAMS and way nicer to code.
        #Use CUSTOM keyword to give custom basis sets, see Crystal manual
        _basis_keys = ['BASISSET']

        #third mandatory block: Hamiltonian and SCF control
        #this needs to be after geometry and basis and this is also the last block
        #so everything else goes in here
        _option_keys = ['OPTIONS']

        def parse(key, value):
            ret = ''

            key = key.upper()
            if isinstance(value, Settings):
                if key in _geom_keys:
                    #geometry block should be available as a list of lines
                    if not isinstance(value, list):
                        raise PlamsError('Geometry block does not support subblocks')

                    #add geometry lines
                    if '_h' in value:
                        ret += '{}\n'.format(value['_h'].upper())
                    i = 1
                    while ('_'+str(i)) in value:
                        ret += parse('',value['_'+str(i)])
                        i += 1

                    for el in value:
                        if not el.startswith('_'):
                            ret += parse(el,value[el])


                elif key in _basis_keys:
                    for subkey, item in value:
                        #we have a CUSTOM basis set
                        if not subkey.upper() == 'CUSTOM':
                            raise PlamsError('BasisSet block only supports subblock CUSTOM')
                        if '_h' in item:
                            ret += '{}\n'.format(item['_h'].upper())
                        i = 1
                        while ('_'+str(i)) in item:
                            ret += parse('',item['_'+str(i)])
                            i += 1

                        for el in item:
                            if not el.startswith('_'):
                                ret += parse('',item[el])
                        ret += 'END\n'

                #option block has no start key
                elif key in _option_keys:
                    if '_h' in value:
                        ret += '{}\n'.format(value['_h'].upper())
                    i = 1
                    while ('_'+str(i)) in value:
                        ret += parse('',value['_'+str(i)])
                        i += 1

                    for el in value:
                        if not el.startswith('_'):
                            ret += parse(el,value[el])
                else:
                    #one line per item
                    ret += '{}\n'.format(key)
                    if '_h' in value:
                        ret += '{}\n'.format(value['_h'].upper())
                    i = 1
                    while ('_'+str(i)) in value:
                        ret += parse('',value['_'+str(i)])
                        i += 1

                    for el in value:
                        if not el.startswith('_'):
                            ret += parse(el,value[el])
                    ret += 'END\n'


            elif isinstance(value, str) and key in [*_geom_keys,*_basis_keys,*_option_keys]:
                ret += '{}\n'.format(value.upper())

            elif isinstance(value, list):
                if not key is '':
                    ret += '{}\n'.format(key)
                for el in value:
                    ret += '{}\n'.format(str(el).upper())

            elif key is '':
                ret += '{}\n'.format(str(value).upper())

            elif value is '' or value is True:
                ret += '{}\n'.format(key)

            elif value is False:
                pass

            else:
                ret += '{}\n{}\n'.format(key, str(value).upper())

            return ret

        inp = ''
        use_molecule = ('ignore_molecule' not in self.settings) or (self.settings.ignore_molecule == False)
        if use_molecule:
            self._parsemol()

        #we need a certain ordering, so make a copy of the settings instance
        #and convert all first-level keys to uppercase
        tmp = Settings()
        for key in self.settings.input:
            tmp[key.upper()] = self.settings.input[key]


        #check for three blocks
        if not any([ x in _geom_keys for x in tmp ]):
            raise PlamsError('One geometry block is necessary for a Crystal Job')
        if not any([ x in _basis_keys for x in tmp ]):
            raise PlamsError('BasisSet block is necessary for a Crystal Job')


        #first title
        inp += '{}\n'.format(self.name)

        #geometry block next
        for item in _geom_keys:
            if item in tmp:
                inp += parse(item, tmp[item])
                #do not close block, it is closed by the BASISSET keyword following
                del tmp[item]

        #basis set block next
        for item in _basis_keys:
            if item in tmp:
                inp += 'BASISSET\n'
                inp += parse(item, tmp[item])
                del tmp[item]

        #everything else now
        for item in tmp:
            inp += parse(item, tmp[item])

        inp += 'END'
        return inp

    def _parsemol(self):
        if 'ase' in Molecule._writeformat:
            #ASE has a write function for Crystal coordinate files, use that if possible
            filename = opj(self.path, 'fort.34')
            #The Crystal IO of ASE only supports filenames, not descirptors right now
            self.molecule.writease(filename)
            self.settings.input.external = True
        else:
            raise PlamsError('Crystal Interface has no builtin Molecule support, install ASE or use function crystalMol2Conf() and set self.settings.ignore_molecule. See Doc for details.')


    def get_runscript(self):
        """
        Run Crystal.
        """
        ret = self._command
        ret += ' < ' + self._filename('inp')
        if self.settings.runscript.stdout_redirect:
            ret += ' >' + self._filename('out')
        ret += '\n\n'
        return ret

    def check(self):
        """
        Look for the normal termination signal in output. Note, that does not mean your calculation was successful!
        """
        termination = self.results.grep_output('TERMINATION')
        return len(termination) == 1



def mol2CrystalConf(molecule):
    """
    Call this function to create a CRYSTAL-type input of your structure:

    Returns a given |Molecule| object as a geomkey and a list of strings that can be used to create a Settings instance for Crystal.

            >>> print(crystalMol2Conf(mol))
            'GEOMKEY', ['0 0 0', '1', 'lattice', 'nAtoms', 'ElementNumber1 X1 Y1 Z1','ElementNumber2 X2 Y2 Z2', ...]

    - IFLAG,IFHR and IFSO are returned as 0,0,0 by default with Symmetry group P1 (number 1). This should allow most calculations to run. The user needs to change them if he wants to take advantage of symmetry.
    - The geometry key is guessed from the number of lattice vectors. For special stuff change it by hand.
    - The number of lattice vectors in the given molecule should correspond to the dimensionality of the system. Do not fill them with zeros or unit vectors, this will result in a 3D-Periodic system with wrong fractional coordinates. So stick with the standard PLAMS way of doing things.
    """
    geomList =  []

    #add lattice keyword
    if len(molecule.lattice) == 0:
         geomKey = 'MOLECULE'
    elif len(molecule.lattice) == 1:
         geomKey = 'POLYMER'
    elif len(molecule.lattice) == 2:
         geomKey = 'SLAB'
    elif len(molecule.lattice) == 3:
         geomKey = 'CRYSTAL'

    #add line for IFLAG,IFHR,IFSO: 0 0 0 always works with P1, only for CRYSTAL
    if geomKey == 'CRYSTAL':
         geomList.append('0 0 0')
    #add a line for space group, assume P1 Symmetry because this always works
    geomList.append('1')

    #add lattice information: vector lengths and angles (if there is any)
    lengths = []
    angles = []
    lattice = molecule.lattice[:]
    for vec in lattice:
        lengths.append(np.linalg.norm(vec))
    if len(lattice) == 2:
        v1_u = lattice[0] / np.linalg.norm(lattice[0])
        v2_u = lattice[1] / np.linalg.norm(lattice[1])
        angle = np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)) / np.pi*180.0
        angles.append(angle)
    elif len(lattice) == 3:
        for first in range(0,3):
            second = first - 1
            third = first - 2
            v1_u = lattice[second] / np.linalg.norm(lattice[second])
            v2_u = lattice[third] / np.linalg.norm(lattice[third])
            angle = np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)) / np.pi*180.0
            angles.append(angle)
    if len(lattice) > 0:
        geomList.append((("{} ")*(len(lengths)+len(angles))).format(*lengths,*angles))

    #add number of atoms
    geomList.append(str(len(molecule)))


    #now add atoms in fractional coordinates or real coords if no lattice
    #transpose lattice
    nDim = len(lattice)
    if nDim > 0:
        if nDim == 1:
            lattice.append(np.array([0.0,1.0,0.0]))
            lattice.append(np.array([0.0,0.0,1.0]))
        elif nDim == 2:
            lattice.append(np.array([0.0,0.0,1.0]))
        #transpose lattice, since PLAMS saves vectors as rows, we need it as columns
        lattice = np.transpose(lattice)
        latticeMatInv = np.linalg.inv(lattice)
        for atom in molecule:
            atomVec = np.dot(latticeMatInv, np.array(atom.coords)).tolist()
            #make the crystal coordinates go from 0 to 1
            for i in range(0,nDim):
                if atomVec[i] < 0.0:
                    while atomVec[i] < 0.0:
                        atomVec[i] += 1.0
                if atomVec[i] > 1.0:
                    while atomVec[i] > 1.0:
                        atomVec[i] -= 1.0
            geomList.append('{:<2}  {:>14f} {:>14f} {:>14f}'.format(atom.atnum,*atomVec))
    else:
        for atom in molecule:
            geomList.append('{:<2}  {:}'.format(atom.atnum,atom.str(symbol=False)))

    return  geomKey, geomList
